########################################################################
# File name: test_service.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import contextlib
import unittest
import unittest.mock

import aioxmpp
import aioxmpp.service as service
import aioxmpp.disco.xso as disco_xso
import aioxmpp.blocking as blocking
import aioxmpp.blocking.xso as blocking_xso

from aioxmpp.utils import namespaces

from aioxmpp.testutils import (
    make_connected_client,
    CoroutineMock,
    run_coroutine,
)

TEST_FROM = aioxmpp.structs.JID.fromstr("foo@bar.example/baz")
TEST_JID1 = aioxmpp.structs.JID.fromstr("bar@bar.example/baz")
TEST_JID2 = aioxmpp.structs.JID.fromstr("baz@bar.example/baz")
TEST_JID3 = aioxmpp.structs.JID.fromstr("quux@bar.example/baz")


class TestBlockingClient(unittest.TestCase):

    def setUp(self):
        self.cc = make_connected_client()
        self.cc.local_jid = TEST_FROM

        self.disco = aioxmpp.DiscoClient(self.cc)
        self.s = blocking.BlockingClient(
            self.cc,
            dependencies={
                aioxmpp.DiscoClient: self.disco,
            }
        )

    def tearDown(self):
        del self.cc
        del self.disco
        del self.s

    def test_is_service(self):
        self.assertTrue(issubclass(
            blocking.BlockingClient,
            aioxmpp.service.Service
        ))

    def test_service_order(self):
        self.assertGreater(
            blocking.BlockingClient,
            aioxmpp.DiscoClient
        )

    def test_get_initial_blocklist_is_depsignal_handler(self):
        self.assertTrue(aioxmpp.service.is_depsignal_handler(
            aioxmpp.Client,
            "before_stream_established",
            self.s._get_initial_blocklist
        ))

    def test_handle_stream_destroyed_is_depsignal_handler(self):
        self.assertTrue(aioxmpp.service.is_depsignal_handler(
            aioxmpp.stream.StanzaStream,
            "on_stream_destroyed",
            self.s.handle_stream_destroyed
        ))

    def test_check_for_blocking(self):
        disco_info = disco_xso.InfoQuery()
        disco_info.features.add(namespaces.xep0191)

        with unittest.mock.patch.object(self.disco, "query_info",
                                        new=CoroutineMock()):
            self.disco.query_info.return_value = disco_info

            run_coroutine(self.s._check_for_blocking())

            self.disco.query_info.assert_called_with(
                TEST_FROM.replace(localpart=None, resource=None)
            )

    def test_check_for_blocking_failure(self):
        disco_info = disco_xso.InfoQuery()
        with unittest.mock.patch.object(self.disco, "query_info",
                                        new=CoroutineMock()):
            self.disco.query_info.return_value = disco_info

            with self.assertRaises(RuntimeError):
                run_coroutine(self.s._check_for_blocking())

            self.disco.query_info.assert_called_with(
                TEST_FROM.replace(localpart=None, resource=None)
            )

    def test_get_initial_blocklist(self):

        with contextlib.ExitStack() as stack:
            stack.enter_context(
                unittest.mock.patch.object(
                    self.s, "_check_for_blocking",
                    new=CoroutineMock()
                )
            )

            stack.enter_context(
                unittest.mock.patch.object(
                    self.cc, "send",
                    new=CoroutineMock()
                )
            )

            handle_initial_blocklist_mock = unittest.mock.Mock()
            self.s.on_initial_blocklist_received.connect(
                handle_initial_blocklist_mock
            )

            BLOCKLIST = [TEST_JID1, TEST_JID2]
            blocklist = blocking_xso.BlockList()
            blocklist.items[:] = BLOCKLIST
            self.cc.send.return_value = blocklist

            result = run_coroutine(self.s._get_initial_blocklist())

            self.assertCountEqual(
                self.s._blocklist,
                BLOCKLIST
            )

            self.assertEqual(len(self.cc.send.mock_calls), 1)
            (_, (arg,), _), = self.cc.send.mock_calls
            self.assertIsInstance(arg, aioxmpp.IQ)
            self.assertEqual(arg.type_, aioxmpp.IQType.GET)
            self.assertIsInstance(arg.payload, blocking_xso.BlockList)
            self.assertEqual(
                len(arg.payload.items),
                0
            )

            self.assertSequenceEqual(
                self.s._check_for_blocking.mock_calls,
                [unittest.mock.call()]
            )

            self.assertSequenceEqual(
                handle_initial_blocklist_mock.mock_calls,
                [
                    unittest.mock.call(
                        frozenset(BLOCKLIST)
                    )
                ]
            )

        self.assertTrue(result)  # so that coroutine doesn’t get disconnected

    def test_get_initial_blocklist_handles_exception(self):
        with contextlib.ExitStack() as stack:
            _check_for_blocking = stack.enter_context(
                unittest.mock.patch.object(
                    self.s, "_check_for_blocking",
                    new=CoroutineMock()
                )
            )
            _check_for_blocking.side_effect = RuntimeError()

            stack.enter_context(
                unittest.mock.patch.object(
                    self.cc, "send",
                    new=CoroutineMock()
                )
            )

            handle_initial_blocklist_mock = unittest.mock.Mock()
            self.s.on_initial_blocklist_received.connect(
                handle_initial_blocklist_mock
            )

            BLOCKLIST = [TEST_JID1, TEST_JID2]
            blocklist = blocking_xso.BlockList()
            blocklist.items[:] = BLOCKLIST
            self.cc.send.return_value = blocklist

            result = run_coroutine(self.s._get_initial_blocklist())

            self.assertIsNone(self.s._blocklist)

            self.cc.send.assert_not_called()
            self.s._check_for_blocking.assert_called_once_with()

        self.assertTrue(result)  # so that coroutine doesn’t get disconnected

    def test_block_jids(self):
        with contextlib.ExitStack() as stack:
            stack.enter_context(
                unittest.mock.patch.object(
                    self.s, "_check_for_blocking",
                    new=CoroutineMock()
                )
            )

            stack.enter_context(
                unittest.mock.patch.object(
                    self.cc, "send",
                    new=CoroutineMock()
                )
            )

            run_coroutine(self.s.block_jids([TEST_JID1]))

            self.assertSequenceEqual(
                self.s._check_for_blocking.mock_calls,
                [unittest.mock.call()]
            )

            self.assertEqual(len(self.cc.send.mock_calls), 1)
            (_, (arg,), _), = self.cc.send.mock_calls

            self.assertIsInstance(arg, aioxmpp.IQ)
            self.assertEqual(arg.type_, aioxmpp.IQType.SET)
            self.assertIsInstance(arg.payload, blocking_xso.BlockCommand)
            self.assertCountEqual(
                arg.payload.items,
                frozenset([TEST_JID1]),
            )

    def test_block_jids_ignore_empty(self):
        with contextlib.ExitStack() as stack:
            stack.enter_context(
                unittest.mock.patch.object(
                    self.s, "_check_for_blocking",
                    new=CoroutineMock()
                )
            )

            stack.enter_context(
                unittest.mock.patch.object(
                    self.cc, "send",
                    new=CoroutineMock()
                )
            )

            run_coroutine(self.s.block_jids([]))

            self.assertSequenceEqual(
                self.s._check_for_blocking.mock_calls,
                [unittest.mock.call()]
            )

            self.assertSequenceEqual(self.cc.send.mock_calls, [])

    def test_unblock_jids(self):
        with contextlib.ExitStack() as stack:
            stack.enter_context(
                unittest.mock.patch.object(
                    self.s, "_check_for_blocking",
                    new=CoroutineMock()
                )
            )

            stack.enter_context(
                unittest.mock.patch.object(
                    self.cc, "send",
                    new=CoroutineMock()
                )
            )

            run_coroutine(self.s.unblock_jids([TEST_JID2]))

            self.assertSequenceEqual(
                self.s._check_for_blocking.mock_calls,
                [unittest.mock.call()]
            )

            self.assertEqual(len(self.cc.send.mock_calls), 1)
            (_, (arg,), _), = self.cc.send.mock_calls

            self.assertIsInstance(arg, aioxmpp.IQ)
            self.assertEqual(arg.type_, aioxmpp.IQType.SET)
            self.assertIsInstance(arg.payload, blocking_xso.UnblockCommand)
            self.assertCountEqual(
                arg.payload.items,
                frozenset([TEST_JID2]),
            )

    def test_unblock_jids_ignore_empty(self):
        with contextlib.ExitStack() as stack:
            stack.enter_context(
                unittest.mock.patch.object(
                    self.s, "_check_for_blocking",
                    new=CoroutineMock()
                )
            )

            stack.enter_context(
                unittest.mock.patch.object(
                    self.cc, "send",
                    new=CoroutineMock()
                )
            )

            run_coroutine(self.s.unblock_jids([]))

            self.assertSequenceEqual(
                self.s._check_for_blocking.mock_calls,
                [unittest.mock.call()]
            )

            self.assertSequenceEqual(self.cc.send.mock_calls, [])

    def test_unblock_all(self):
        with contextlib.ExitStack() as stack:
            stack.enter_context(
                unittest.mock.patch.object(
                    self.s, "_check_for_blocking",
                    new=CoroutineMock()
                )
            )

            stack.enter_context(
                unittest.mock.patch.object(
                    self.cc, "send",
                    new=CoroutineMock()
                )
            )

            run_coroutine(self.s.unblock_all())

            self.assertSequenceEqual(
                self.s._check_for_blocking.mock_calls,
                [unittest.mock.call()]
            )

    def test_handle_block_push_is_iq_handler(self):
        service.is_iq_handler(aioxmpp.IQType.SET,
                              blocking_xso.BlockCommand,
                              self.s.handle_block_push)

    def test_handle_block_push(self):
        handle_block = unittest.mock.Mock()
        handle_unblock = unittest.mock.Mock()

        self.s.on_jids_blocked.connect(
            handle_block
        )

        self.s.on_jids_unblocked.connect(
            handle_unblock
        )

        self.s._blocklist = frozenset([TEST_JID1])

        block = blocking_xso.BlockCommand()
        block.items[:] = [TEST_JID2]
        iq = aioxmpp.IQ(
            type_=aioxmpp.IQType.SET,
            payload=block,
        )

        run_coroutine(self.s.handle_block_push(iq))

        self.assertEqual(
            self.s._blocklist,
            frozenset([TEST_JID1, TEST_JID2])
        )

        self.assertEqual(
            handle_block.mock_calls,
            [
                unittest.mock.call(
                    frozenset([TEST_JID2])
                )
            ]
        )

        handle_unblock.assert_not_called()

    def test_handle_unblock_push_is_iq_handler(self):
        service.is_iq_handler(aioxmpp.IQType.SET,
                              blocking_xso.UnblockCommand,
                              self.s.handle_unblock_push)

    def test_handle_unblock_push(self):
        handle_block = unittest.mock.Mock()
        handle_unblock = unittest.mock.Mock()

        self.s.on_jids_blocked.connect(
            handle_block
        )

        self.s.on_jids_unblocked.connect(
            handle_unblock
        )

        self.s._blocklist = frozenset([TEST_JID1, TEST_JID2])

        block = blocking_xso.UnblockCommand()
        block.items[:] = [TEST_JID2]
        iq = aioxmpp.IQ(
            type_=aioxmpp.IQType.SET,
            payload=block,
        )

        run_coroutine(self.s.handle_unblock_push(iq))

        self.assertEqual(
            self.s._blocklist,
            frozenset([TEST_JID1])
        )

        self.assertEqual(
            handle_unblock.mock_calls,
            [
                unittest.mock.call(
                    frozenset([TEST_JID2])
                )
            ]
        )

        handle_block.assert_not_called()

    def test_handle_unblock_push_all(self):
        handle_block = unittest.mock.Mock()
        handle_unblock = unittest.mock.Mock()

        self.s.on_jids_blocked.connect(
            handle_block
        )

        self.s.on_jids_unblocked.connect(
            handle_unblock
        )

        self.s._blocklist = frozenset([TEST_JID1, TEST_JID2])

        block = blocking_xso.UnblockCommand()
        iq = aioxmpp.IQ(
            type_=aioxmpp.IQType.SET,
            payload=block,
        )

        run_coroutine(self.s.handle_unblock_push(iq))

        self.assertEqual(
            handle_unblock.mock_calls,
            [
                unittest.mock.call(
                    frozenset([TEST_JID1, TEST_JID2])
                )
            ]
        )

        handle_block.assert_not_called()



########################################################################
# File name: hashes.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
"""
:mod:`~aioxmpp.hashes` --- Hash Functions for use with XMPP (:xep:`300`)
########################################################################

:xep:`300` consolidates the use of hash functions and their digests in XMPP.
Identifiers (usually called `algo`) are defined to refer to specific
implementations and parametrisations of hashes (:func:`hash_from_algo`,
:func:`algo_of_hash`) and there is a defined XML format for carrying hash
digests (:class:`Hash`).

This allows other extensions to easily embed hash digests in their protocols
(:class:`HashesParent`).

.. note::

    Compliance with :xep:`300` depends on your build of Python and possibly
    OpenSSL. Version 0.5.1 of :xep:`300` requires support of SHA3 and BLAKE2b,
    which was only introduced in Python 3.6.

Utilities for Working with Hash Algorithm Identifiers
=====================================================

.. autofunction:: hash_from_algo

.. autofunction:: algo_of_hash

.. data:: default_hash_algorithms

    A set of `algo` values which consists of hash functions matching the
    following criteria:

    * They are specified as ``MUST`` or ``SHOULD`` in the supported version of
      :xep:`300`.
    * They are supported by :mod:`hashlib`.
    * Only one function from each matching family is selected. If multiple
      functions apply, ``MUST`` is preferred over ``SHOULD``.

    The set thus varies based on the build of Python and possibly OpenSSL. The
    algorithms in the set are guaranteed to return a valid hash implementation
    when passed to :func:`~aioxmpp.misc.hash_from_algo`.

    In a fully compliant build, this set consists of ``sha-256``, ``sha3-256``
    and ``blake2b-256``.

XSOs
====

.. autoclass:: Hash

.. autoclass:: HashesParent()
"""
import hashlib

import aioxmpp.xso as xso

from aioxmpp.utils import namespaces

namespaces.xep0300_hashes2 = "urn:xmpp:hashes:2"


_HASH_ALGO_MAPPING = [
    ("md2", (False, ("md2", (), {}))),
    ("md4", (False, ("md4", (), {}))),
    ("md5", (False, ("md5", (), {}))),
    ("sha-1", (True, ("sha1", (), {}))),
    ("sha-224", (True, ("sha224", (), {}))),
    ("sha-256", (True, ("sha256", (), {}))),
    ("sha-384", (True, ("sha384", (), {}))),
    ("sha-512", (True, ("sha512", (), {}))),
    ("sha3-256", (True, ("sha3_256", (), {}))),
    ("sha3-512", (True, ("sha3_512", (), {}))),
    ("blake2b-256", (True, ("blake2b", (), {"digest_size": 32}))),
    ("blake2b-512", (True, ("blake2b", (), {"digest_size": 64}))),
]


_HASH_ALGO_MAP = dict(_HASH_ALGO_MAPPING)
_HASH_ALGO_REVERSE_MAP = {
    fun_name: (enabled, algo)
    for algo, (enabled, (fun_name, fun_args, fun_kwargs)) in _HASH_ALGO_MAPPING
    if not fun_args and not fun_kwargs
}


def is_algo_supported(algo):
    try:
        enabled, (fun_name, _, _) = _HASH_ALGO_MAP[algo]
    except KeyError:
        return False

    return enabled and hasattr(hashlib, fun_name)


def hash_from_algo(algo):
    """
    Return a :mod:`hashlib` hash given the :xep:`300` `algo`.

    :param algo: The algorithm identifier as defined in :xep:`300`.
    :type algo: :class:`str`
    :raises NotImplementedError: if the hash algortihm is not supported by
        :mod:`hashlib`.
    :raises ValueError: if the hash algorithm MUST NOT be supported.
    :return: A hash object from :mod:`hashlib` or compatible.

    If the `algo` is not supported by the :mod:`hashlib` module,
    :class:`NotImplementedError` is raised.
    """

    try:
        enabled, (fun_name, fun_args, fun_kwargs) = _HASH_ALGO_MAP[algo]
    except KeyError as exc:
        raise NotImplementedError(
            "hash algorithm {!r} unknown".format(algo)
        ) from None

    if not enabled:
        raise ValueError(
            "support of {} in XMPP is forbidden".format(algo)
        )

    try:
        fun = getattr(hashlib, fun_name)
    except AttributeError as exc:
        raise NotImplementedError(
            "{} not supported by hashlib".format(algo)
        ) from exc

    return fun(*fun_args, **fun_kwargs)


def algo_of_hash(h):
    """
    Return a :xep:`300` `algo` from a given :mod:`hashlib` hash.

    :param h: Hash object from :mod:`hashlib`.
    :raises ValueError: if `h` does not have a defined `algo` value.
    :raises ValueError: if the hash function MUST NOT be supported.
    :return: The `algo` value for the given hash.
    :rtype: :class:`str`

    .. warning::

        Use with caution for :func:`hashlib.blake2b` hashes.
        :func:`algo_of_hash` cannot safely determine whether blake2b was
        initialised with a salt, personality, key or other non-default
        :xep:`300` mode.

        In such a case, the return value will be the matching ``blake2b-*``
        `algo`, but the digest will not be compatible with the results of other
        implementations.

    """
    try:
        enabled, algo = _HASH_ALGO_REVERSE_MAP[h.name]
    except KeyError:
        pass
    else:
        if not enabled:
            raise ValueError("support of {} in XMPP is forbidden".format(
                algo
            ))
        return algo

    if h.name == "blake2b":
        return "blake2b-{}".format(h.digest_size * 8)

    raise ValueError(
        "unknown hash implementation: {!r}".format(h)
    )


class Hash(xso.XSO):
    """
    Represent a single hash digest.

    .. attribute:: algo

        The hash algorithm used. The name is as specified in :xep:`300`.

    .. attribute:: digest

        The digest as :class:`bytes`.

    """

    TAG = namespaces.xep0300_hashes2, "hash"

    algo = xso.Attr(
        "algo",
    )

    digest = xso.Text(
        type_=xso.Base64Binary()
    )

    def __init__(self, algo, digest):
        super().__init__()
        self.algo = algo
        self.digest = digest

    def get_impl(self):
        """
        Return a new :mod:`hashlib` hash for the :attr:`algo` set on this
        object.

        See :func:`hash_from_algo` for details and exceptions.
        """
        return hash_from_algo(self.algo)


class HashType(xso.AbstractElementType):
    @classmethod
    def get_xso_types(cls):
        return [Hash]

    def unpack(self, obj):
        return obj.algo, obj.digest

    def pack(self, pair):
        return Hash(*pair)


class HashesParent(xso.XSO):
    """
    Mix-in class for XSOs which use :class:`Hash` children.

    .. attribute:: digests

        A mapping which maps from the :attr:`Hash.algo` to the
        :attr:`Hash.digest`.
    """

    digests = xso.ChildValueMap(
        type_=HashType(),
    )


default_hash_algorithms = {
    algo
    for algo in ["sha-256", "sha3-256", "blake2b-256"]
    if is_algo_supported(algo)
}



########################################################################
# File name: test_dispatcher.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import unittest

import aioxmpp.callbacks
import aioxmpp.carbons.xso as carbons_xso
import aioxmpp.muc
import aioxmpp.im.dispatcher as dispatcher
import aioxmpp.service
import aioxmpp.stream

from aioxmpp.utils import namespaces

from aioxmpp.testutils import (
    make_connected_client,
    CoroutineMock,
    run_coroutine,
)


TEST_LOCAL = aioxmpp.JID.fromstr("foo@service.example")
TEST_PEER = aioxmpp.JID.fromstr("bar@service.example")


class TestIMDispatcher(unittest.TestCase):
    def setUp(self):
        self.cc = make_connected_client()
        self.cc.local_jid = TEST_LOCAL.replace(resource="we")
        self.disco_client = aioxmpp.DiscoClient(self.cc)
        self.carbons = aioxmpp.CarbonsClient(self.cc, dependencies={
            aioxmpp.DiscoClient: self.disco_client,
        })
        self.s = dispatcher.IMDispatcher(self.cc, dependencies={
            aioxmpp.CarbonsClient: self.carbons,
        })
        self.listener = unittest.mock.Mock()

        for filter_ in ["message_filter", "presence_filter"]:
            cb = getattr(self.listener, filter_)
            cb.side_effect = lambda x, *args, **kwargs: x
            filter_chain = getattr(self.s, filter_)
            filter_chain.register(cb, 0)

    def tearDown(self):
        del self.s
        del self.cc

    def test_message_filter_is_filter(self):
        self.assertIsInstance(
            self.s.message_filter,
            aioxmpp.callbacks.Filter,
        )

    def test_presence_filter_is_filter(self):
        self.assertIsInstance(
            self.s.presence_filter,
            aioxmpp.callbacks.Filter,
        )

    def test_orders_after_simple_presence_dispatcher(self):
        self.assertIn(
            aioxmpp.dispatcher.SimplePresenceDispatcher,
            dispatcher.IMDispatcher.ORDER_AFTER,
        )

    def test_depends_on_carbons(self):
        self.assertIn(
            aioxmpp.CarbonsClient,
            dispatcher.IMDispatcher.ORDER_AFTER,
        )

    def test_dispatcher_connects_to_before_stream_established(self):
        self.assertTrue(
            aioxmpp.service.is_depsignal_handler(
                aioxmpp.Client,
                "before_stream_established",
                dispatcher.IMDispatcher.enable_carbons,
            )
        )

    def test_dispatch_message_listens_to_on_message_received(self):
        self.assertTrue(
            aioxmpp.service.is_depsignal_handler(
                aioxmpp.stream.StanzaStream,
                "on_message_received",
                dispatcher.IMDispatcher.dispatch_message,
            )
        )

    def test_dispatch_presence_listens_to_on_presence_received(self):
        self.assertTrue(
            aioxmpp.service.is_depsignal_handler(
                aioxmpp.stream.StanzaStream,
                "on_presence_received",
                dispatcher.IMDispatcher.dispatch_presence,
            )
        )

    def test_dispatch_presences(self):
        types = [
            (aioxmpp.PresenceType.AVAILABLE, True),
            (aioxmpp.PresenceType.UNAVAILABLE, True),
            (aioxmpp.PresenceType.ERROR, True),
            (aioxmpp.PresenceType.SUBSCRIBE, False),
            (aioxmpp.PresenceType.UNSUBSCRIBE, False),
            (aioxmpp.PresenceType.SUBSCRIBED, False),
            (aioxmpp.PresenceType.UNSUBSCRIBED, False),
        ]

        for type_, dispatch in types:
            pres = aioxmpp.Presence(
                type_=type_,
                from_=TEST_PEER,
                to=TEST_LOCAL,
            )

        self.s.dispatch_presence(pres)

        self.listener.presence_filter.assert_called_once_with(
            pres,
            TEST_PEER,
            False,
        )

    def test_dispatch_simple_messages(self):
        msg = aioxmpp.Message(
            type_=aioxmpp.MessageType.CHAT,
            from_=TEST_PEER,
            to=TEST_LOCAL,
        )

        self.s.dispatch_message(msg)

        self.listener.message_filter.assert_called_once_with(
            msg,
            TEST_PEER,
            False,
            dispatcher.MessageSource.STREAM,
        )

    def test_dispatch_sent_messages(self):
        msg = aioxmpp.Message(
            type_=aioxmpp.MessageType.CHAT,
            from_=TEST_LOCAL,
            to=TEST_PEER,
        )

        self.s.dispatch_message(msg, sent=True)

        self.listener.message_filter.assert_called_once_with(
            msg,
            TEST_PEER,
            True,
            dispatcher.MessageSource.STREAM,
        )

    def test_dispatch_unpacks_received_carbon(self):
        msg = aioxmpp.Message(
            type_=aioxmpp.MessageType.CHAT,
            from_=TEST_PEER,
            to=TEST_LOCAL,
        )

        wrapper = aioxmpp.Message(
            type_=msg.type_,
            from_=TEST_LOCAL.bare(),
            to=TEST_LOCAL,
        )
        wrapper.xep0280_received = carbons_xso.Received()
        wrapper.xep0280_received.stanza = msg

        self.s.dispatch_message(wrapper)

        self.listener.message_filter.assert_called_once_with(
            msg,
            TEST_PEER,
            False,
            dispatcher.MessageSource.CARBONS,
        )

    def test_dispatch_drops_received_carbon_with_incorrect_from(self):
        msg = aioxmpp.Message(
            type_=aioxmpp.MessageType.CHAT,
            from_=TEST_LOCAL,
            to=TEST_PEER,
        )

        wrapper = aioxmpp.Message(
            type_=msg.type_,
            from_=TEST_PEER,
            to=TEST_LOCAL,
        )
        wrapper.xep0280_received = carbons_xso.Received()
        wrapper.xep0280_received.stanza = msg

        self.s.dispatch_message(wrapper)

        self.listener.message_filter.assert_not_called()

    def test_dispatch_unpacks_sent_carbon(self):
        msg = aioxmpp.Message(
            type_=aioxmpp.MessageType.CHAT,
            from_=TEST_LOCAL.replace(resource="other"),
            to=TEST_PEER,
        )

        wrapper = aioxmpp.Message(
            type_=msg.type_,
            from_=TEST_LOCAL.bare(),
            to=TEST_LOCAL,
        )
        wrapper.xep0280_sent = carbons_xso.Sent()
        wrapper.xep0280_sent.stanza = msg

        self.s.dispatch_message(wrapper)

        self.listener.message_filter.assert_called_once_with(
            msg,
            TEST_PEER,
            True,
            dispatcher.MessageSource.CARBONS,
        )

    def test_dispatch_drops_sent_carbon_with_incorrect_from(self):
        msg = aioxmpp.Message(
            type_=aioxmpp.MessageType.CHAT,
            from_=TEST_LOCAL,
            to=TEST_PEER,
        )

        wrapper = aioxmpp.Message(
            type_=msg.type_,
            from_=TEST_PEER,
            to=TEST_LOCAL,
        )
        wrapper.xep0280_sent = carbons_xso.Sent()
        wrapper.xep0280_sent.stanza = msg

        self.s.dispatch_message(wrapper)

        self.listener.message_filter.assert_not_called()

    def test_enable_carbons_enables_carbons(self):
        with unittest.mock.patch.object(
                self.carbons,
                "enable",
                new=CoroutineMock()) as enable:
            result = run_coroutine(self.s.enable_carbons())

        enable.assert_called_once_with()
        self.assertTrue(result)

    def test_enable_carbons_does_not_swallow_random_exception(self):
        class FooException(Exception):
            pass

        with unittest.mock.patch.object(
                self.carbons,
                "enable",
                new=CoroutineMock()) as enable:
            enable.side_effect = FooException()
            with self.assertRaises(FooException):
                run_coroutine(self.s.enable_carbons())

        enable.assert_called_once_with()

    def test_enable_carbons_ignores_RuntimeError_from_enable(self):
        with unittest.mock.patch.object(
                self.carbons,
                "enable",
                new=CoroutineMock()) as enable:
            enable.side_effect = RuntimeError()
            run_coroutine(self.s.enable_carbons())

        enable.assert_called_once_with()

    def test_enable_carbons_ignores_XMPPError_from_enable(self):
        with unittest.mock.patch.object(
                self.carbons,
                "enable",
                new=CoroutineMock()) as enable:
            enable.side_effect = aioxmpp.errors.XMPPError(
                aioxmpp.ErrorCondition.REMOTE_SERVER_NOT_FOUND
            )
            run_coroutine(self.s.enable_carbons())

        enable.assert_called_once_with()



########################################################################
# File name: carbons_sniffer.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import asyncio

import aioxmpp

from framework import Example, exec_example


class CarbonsSniffer(Example):
    def make_simple_client(self):
        client = super().make_simple_client()
        self.carbons = client.summon(aioxmpp.CarbonsClient)
        return client

    def _format_message(self, message):
        parts = []
        parts.append(str(message))
        if message.body:
            parts.append("text: {}".format(message.body))
        else:
            parts.append("other")

        return "; ".join(parts)

    def _message_filter(self, message):
        if     (message.from_ != self.client.local_jid.bare() and
                message.from_ is not None):
            return

        if message.xep0280_sent is not None:
            print("SENT: {}".format(self._format_message(
                message.xep0280_sent.stanza
            )))

        elif message.xep0280_received is not None:
            print("RECV: {}".format(self._format_message(
                message.xep0280_received.stanza
            )))


    @asyncio.coroutine
    def run_example(self):
        self.stop_event = self.make_sigint_event()
        yield from super().run_example()

    @asyncio.coroutine
    def run_simple_example(self):
        filterchain = self.client.stream.app_inbound_message_filter
        with filterchain.context_register(self._message_filter):
            print("enabling carbons")
            yield from self.carbons.enable()
            print("carbons enabled! sniffing ... (hit Ctrl+C to stop)")

            yield from self.stop_event.wait()


if __name__ == "__main__":
    exec_example(CarbonsSniffer())



########################################################################
# File name: xso.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import aioxmpp
import aioxmpp.xso as xso

from aioxmpp.utils import namespaces

namespaces.xep0184_receipts = "urn:xmpp:receipts"


aioxmpp.Message.xep0184_request_receipt = xso.ChildFlag(
    (namespaces.xep0184_receipts, "request"),
)


class Received(xso.XSO):
    TAG = namespaces.xep0184_receipts, "received"

    message_id = xso.Attr(
        "id",
    )

    def __init__(self, message_id):
        super().__init__()
        self.message_id = message_id


aioxmpp.Message.xep0184_received = xso.Child(
    [
        Received,
    ]
)



########################################################################
# File name: server_info.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import asyncio
import itertools

import aioxmpp.disco

from framework import Example, exec_example


class ServerInfo(Example):
    @asyncio.coroutine
    def run_simple_example(self):
        disco = self.client.summon(aioxmpp.DiscoClient)
        try:
            info = yield from disco.query_info(
                self.g_jid.replace(resource=None, localpart=None),
                timeout=10
            )
        except Exception as exc:
            print("could not get info: ")
            print("{}: {}".format(type(exc).__name__, exc))
            raise

        print("features:")
        for feature in info.features:
            print("  {!r}".format(feature))

        print("identities:")
        identities = list(info.identities)

        def identity_key(ident):
            return (ident.category, ident.type_)

        identities.sort(key=identity_key)
        for (category, type_), identities in (
                itertools.groupby(info.identities, identity_key)):
            print("  category={!r} type={!r}".format(category, type_))
            subidentities = list(identities)
            subidentities.sort(key=lambda ident: ident.lang)
            for identity in subidentities:
                print("    [{}] {!r}".format(identity.lang, identity.name))


if __name__ == "__main__":
    exec_example(ServerInfo())



########################################################################
# File name: new-xml-concept.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
"""
New XML concept
===============

The idea is to have stanza classes describe what their contents shall be and how
these are serialized/deserialized to/from XML.

.. class:: StanzaModel.Attr(name=<autodected>,
                            type_=StanzaModel.String,
                            required=False,
                            restrict=None,
                            default=None)

   *required* must be either a :data:`bool` or a callable taking one
   argument. If it is a callable, during validation, it is passed the instance
   which is currently being validated. The instance has been fully populated at
   that time, but not yet validated. The callable shall return a boolean value
   indicating whether the attribute is required or not.

   *type_* must be a :class:`StanzaModel.Type`. It is used to parse the string
   contents of the attribute and convert it to a python type. To impose
   additional restrictions on the type, use the *restrict* argument.

   *restrict* must be a :class:`StanzaModel.Restriction` or :data:`None`, which
   is used to *validate* the value extracted using the *type_* object. Passing
   :data:`None` indicates that all values which are accepted by the *type_* are
   valid.

   *default* is used if the attribute is absent and *required* evaluates to
   :data:`False`. *default* is not checked against either the *type_* or the
   *restrict* objects.

.. class:: StanzaModel.Child(match="*",
                             n=1,
                             required=False)

   *required* works as in :class:`StanzaModel.Attr`.

   *n* must be either a positive integer (indicating the exact amount of
   elements which may occur) or a tuple of two positive integers. If it is a
   tuple, the first value may also be ``0`` and the second value may be
   :data:`None`, in which case no upper bound is specified.

   *match* must be either a string expression which can be used with
   :meth:`lxml.etree._Element.iterchildren` or a :class:`StanzaObject`
   class. In the latter case, the :attr:`StanzaObject.TAG` attribute is used as
   a matcher. It restricts the set of elements to be considered for this
   attribute.

   If *n* is a tuple or not equal to ``1``, the elements are stored in a
   list. Otherwise, the element must be accessible directly.

.. class:: StanzaModel.Text(type_=StanzaModel.String,
                            required=False,
                            restrict=None,
                            default=None)

   All arguments work like those applying to :class:`StanzaModel.Attr`.

.. note::

   Mixing :class:`StanzaModel.Text` and :class:`StanzaModel.Child` descriptors
   on the same :class:`StanzaObject` won’t end well. The semantics of
   :class:`StanzaModel.Text` is undefined in that case.

"""

class IQ(metaclass=StanzaObject):
    TAG = "{jabber:client}iq"

    type_ = xml_attr(
        name="type",
        type_=StanzaModel.enum("set", "get", "error", "result"),
        required=True
    )

    data = xml_child(
        match="*",
        n=1,
        required=lambda instance: instance.type_ in {"set", "get"}
    )



########################################################################
# File name: xso.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import collections.abc
import enum

import aioxmpp.stanza
import aioxmpp.forms
import aioxmpp.xso as xso

from aioxmpp.utils import namespaces

namespaces.xep0050_commands = "http://jabber.org/protocol/commands"


class NoteType(enum.Enum):
    INFO = "info"
    WARN = "warn"
    ERROR = "error"


class ActionType(enum.Enum):
    NEXT = "next"
    EXECUTE = "execute"
    PREV = "prev"
    CANCEL = "cancel"
    COMPLETE = "complete"


class CommandStatus(enum.Enum):
    """
    Describes the status a command execution is in.

    .. attribute:: EXECUTING

       The command is being executed.

    .. attribute:: COMPLETED

       The command has been completed.

    .. attribute:: CANCELED

       The command has been canceled.
    """

    EXECUTING = "executing"
    COMPLETED = "completed"
    CANCELED = "canceled"


class Note(xso.XSO):
    TAG = (namespaces.xep0050_commands, "note")

    body = xso.Text(
        default=None,
    )

    type_ = xso.Attr(
        "type",
        type_=xso.EnumCDataType(
            NoteType,
        ),
        default=NoteType.INFO,
    )

    def __init__(self, type_, body):
        super().__init__()
        self.type_ = type_
        self.body = body


class Actions(xso.XSO):
    TAG = (namespaces.xep0050_commands, "actions")

    next_is_allowed = xso.ChildFlag(
        (namespaces.xep0050_commands, "next"),
    )

    prev_is_allowed = xso.ChildFlag(
        (namespaces.xep0050_commands, "prev"),
    )

    complete_is_allowed = xso.ChildFlag(
        (namespaces.xep0050_commands, "complete"),
    )

    execute = xso.Attr(
        "execute",
        type_=xso.EnumCDataType(ActionType),
        validator=xso.RestrictToSet({
            ActionType.NEXT,
            ActionType.PREV,
            ActionType.COMPLETE,
        }),
        default=None,
    )

    @property
    def allowed_actions(self):
        result = [ActionType.EXECUTE, ActionType.CANCEL]
        if self.prev_is_allowed:
            result.append(ActionType.PREV)
        if self.next_is_allowed:
            result.append(ActionType.NEXT)
        if self.complete_is_allowed:
            result.append(ActionType.COMPLETE)
        return frozenset(result)

    @allowed_actions.setter
    def allowed_actions(self, values):
        values = frozenset(values)
        if ActionType.EXECUTE not in values:
            raise ValueError("EXECUTE must always be allowed")
        if ActionType.CANCEL not in values:
            raise ValueError("CANCEL must always be allowed")
        self.prev_is_allowed = ActionType.PREV in values
        self.next_is_allowed = ActionType.NEXT in values
        self.complete_is_allowed = ActionType.COMPLETE in values


@aioxmpp.IQ.as_payload_class
class Command(xso.XSO):
    TAG = (namespaces.xep0050_commands, "command")

    actions = xso.Child([Actions])

    notes = xso.ChildList([Note])

    action = xso.Attr(
        "action",
        type_=xso.EnumCDataType(ActionType),
        default=ActionType.EXECUTE,
    )

    status = xso.Attr(
        "status",
        type_=xso.EnumCDataType(CommandStatus),
        default=None,
    )

    sessionid = xso.Attr(
        "sessionid",
        default=None,
    )

    node = xso.Attr(
        "node",
    )

    payload = xso.ChildList([
        aioxmpp.forms.Data,
    ])

    def __init__(self, node, *,
                 action=ActionType.EXECUTE,
                 status=None,
                 sessionid=None,
                 payload=[],
                 notes=[],
                 actions=None):
        super().__init__()
        self.node = node
        self.action = action
        self.status = status
        self.sessionid = sessionid
        if not isinstance(payload, collections.abc.Iterable):
            self.payload[:] = [payload]
        else:
            self.payload[:] = payload
        self.notes[:] = notes
        self.actions = actions

    @property
    def first_payload(self):
        try:
            return self.payload[0]
        except IndexError:
            return


MalformedAction = aioxmpp.stanza.make_application_error(
    "MalformedAction",
    (namespaces.xep0050_commands, "malformed-action"),
)

BadAction = aioxmpp.stanza.make_application_error(
    "BadAction",
    (namespaces.xep0050_commands, "bad-action"),
)

BadLocale = aioxmpp.stanza.make_application_error(
    "BadLocale",
    (namespaces.xep0050_commands, "bad-locale"),
)

BadPayload = aioxmpp.stanza.make_application_error(
    "BadPayload",
    (namespaces.xep0050_commands, "bad-payload"),
)

BadSessionID = aioxmpp.stanza.make_application_error(
    "BadSessionID",
    (namespaces.xep0050_commands, "bad-sessionid"),
)

SessionExpired = aioxmpp.stanza.make_application_error(
    "SessionExpired",
    (namespaces.xep0050_commands, "session-expired"),
)



########################################################################
# File name: test_xso.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import unittest

import aioxmpp.structs
import aioxmpp.bookmarks.xso as bookmark_xso
import aioxmpp.xso as xso

from aioxmpp.utils import namespaces

EXAMPLE_JID1 = aioxmpp.JID.fromstr("coven@conference.shakespeare.lit")
EXAMPLE_JID2 = aioxmpp.JID.fromstr("open_field@conference.shakespeare.lit")


class TestNamespace(unittest.TestCase):

    def test_namespace(self):
        self.assertEqual(namespaces.xep0048, "storage:bookmarks")


class TestStorage(unittest.TestCase):

    def test_is_xso(self):
        self.assertTrue(issubclass(bookmark_xso.Storage, xso.XSO))

    def test_tag(self):
        self.assertEqual(
            bookmark_xso.Storage.TAG,
            (namespaces.xep0048, "storage")
        )


class TestConference(unittest.TestCase):

    def test_is_xso(self):
        self.assertTrue(issubclass(bookmark_xso.Conference, xso.XSO))

    def test_tag(self):
        self.assertEqual(
            bookmark_xso.Conference.TAG,
            (namespaces.xep0048, "conference")
        )

    def test_init(self):
        conference = bookmark_xso.Conference("Coven", EXAMPLE_JID1)
        self.assertEqual(conference.name, "Coven")
        self.assertEqual(conference.jid, EXAMPLE_JID1)
        self.assertEqual(conference.autojoin, False)
        self.assertEqual(conference.nick, None)
        self.assertEqual(conference.password, None)

        conference = bookmark_xso.Conference(
            "Coven", EXAMPLE_JID1,
            autojoin=True, nick="First Witch",
            password="h3c473"
        )
        self.assertEqual(conference.name, "Coven")
        self.assertEqual(conference.jid, EXAMPLE_JID1)
        self.assertEqual(conference.autojoin, True)
        self.assertEqual(conference.nick, "First Witch")
        self.assertEqual(conference.password, "h3c473")

    def test_eq(self):
        conf = bookmark_xso.Conference("Coven", EXAMPLE_JID1)
        conf1 = bookmark_xso.Conference("Coven", EXAMPLE_JID1)
        conf2 = bookmark_xso.Conference(
            "Coven1", EXAMPLE_JID1,
            autojoin=True, nick="First Witch",
            password="h3c473"
        )
        conf3 = bookmark_xso.Conference(
            "Coven", EXAMPLE_JID2,
            autojoin=True, nick="First Witch",
            password="h3c473"
        )
        conf4 = bookmark_xso.Conference(
            "Coven", EXAMPLE_JID1,
            nick="First Witch",
            password="h3c473"
        )
        conf5 = bookmark_xso.Conference(
            "Coven", EXAMPLE_JID1,
            autojoin=True,
            password="h3c473"
        )
        conf6 = bookmark_xso.Conference(
            "Coven", EXAMPLE_JID1,
            autojoin=True, nick="First Witch"
        )
        url = bookmark_xso.URL(
            "Coven", "xmpp://coven@conference.shakespeare.lit"
        )
        self.assertEqual(conf, conf)
        self.assertEqual(conf, conf1)
        self.assertNotEqual(conf, conf2)
        self.assertNotEqual(conf, conf3)
        self.assertNotEqual(conf, conf4)
        self.assertNotEqual(conf, conf5)
        self.assertNotEqual(conf, conf6)
        self.assertNotEqual(conf, url)


class TestURL(unittest.TestCase):

    def test_is_xso(self):
        self.assertTrue(issubclass(bookmark_xso.URL, xso.XSO))

    def test_tag(self):
        self.assertEqual(
            bookmark_xso.URL.TAG,
            (namespaces.xep0048, "url")
        )

    def test_init(self):
        url = bookmark_xso.URL("Url1", "http://example.com")
        self.assertEqual(url.name, "Url1")
        self.assertEqual(url.url, "http://example.com")

    def test_eq(self):
        url = bookmark_xso.URL("Url", "http://example.com")
        url1 = bookmark_xso.URL("Url", "http://example.com")
        url2 = bookmark_xso.URL("Url1", "http://example.com")
        url3 = bookmark_xso.URL("Url", "http://example.com/foo")
        conf = bookmark_xso.Conference("Coven", EXAMPLE_JID1)

        self.assertEqual(url, url)
        self.assertEqual(url, url1)
        self.assertNotEqual(url, url2)
        self.assertNotEqual(url, url3)
        self.assertNotEqual(url, conf)


@bookmark_xso.as_bookmark_class
class CustomBookmark(bookmark_xso.Bookmark):
    TAG = ("urn:example:bookmark", "bookmark")

    def __init__(self, name, contents):
        self.name = name
        self.contents = contents

    name = aioxmpp.xso.Attr("name")
    contents = aioxmpp.xso.Attr("contents")

    @property
    def primary(self):
        return self.contents

    @property
    def secondary(self):
        return (self.name,)


class TestCustomBookmark(unittest.TestCase):

    def test_registered(self):
        self.assertIn(
            CustomBookmark, bookmark_xso.Storage.bookmarks._classes
        )

    def test_non_Bookmarks_fail(self):
        with self.assertRaises(TypeError):

            @bookmark_xso.as_bookmark_class
            class CustomBookmark(aioxmpp.xso.XSO):
                TAG = ("urn:example:bookmark", "bookmark")

                def __init__(self, name, contents):
                    self.name = name
                    self.contents = contents

                name = aioxmpp.xso.Attr("name")
                contents = aioxmpp.xso.Attr("contents")

                @property
                def primary(self):
                    return self.contents

                @property
                def secondary(self):
                    return (self.name,)



########################################################################
# File name: __init__.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
"""
Version information
###################

There are two ways to obtain the imported version of the :mod:`aioxmpp`
package:

.. autodata:: __version__

.. data:: version

   Alias of :data:`__version__`.

.. autodata:: version_info

.. _api-aioxmpp-services:

Overview of Services
####################

.. autosummary::
    :nosignatures:

    aioxmpp.AdHocClient
    aioxmpp.AvatarService
    aioxmpp.BlockingClient
    aioxmpp.BookmarkClient
    aioxmpp.CarbonsClient
    aioxmpp.DiscoClient
    aioxmpp.DiscoServer
    aioxmpp.EntityCapsService
    aioxmpp.MUCClient
    aioxmpp.PingService
    aioxmpp.PresenceClient
    aioxmpp.PresenceServer
    aioxmpp.PEPClient
    aioxmpp.RosterClient
    aioxmpp.VersionServer

Shorthands
##########

.. function:: make_security_layer

   Alias of :func:`aioxmpp.security_layer.make`.

"""
from ._version import version_info, __version__, version

#: The imported :mod:`aioxmpp` version as a tuple.
#:
#: The components of the tuple are, in order: `major version`, `minor version`,
#: `patch level`, and `pre-release identifier`.
#:
#: .. seealso::
#:
#:    :ref:`api-stability`
version_info = version_info

#: The imported :mod:`aioxmpp` version as a string.
#:
#: The version number is dot-separated; in pre-release or development versions,
#: the version number is followed by a hypen-separated pre-release identifier.
#:
#: .. seealso::
#:
#:    :ref:`api-stability`
__version__ = __version__

# XXX: ^ this is a hack to make Sphinx find the docs. We could also be using
# .. data instead of .. autodata, but that has the downside that the actual
# version number isn’t printed in the docs (without additional maintenance
# cost).

import asyncio # NOQA
# Adds fallback if asyncio version does not provide an ensure_future function.
if not hasattr(asyncio, "ensure_future"):
    asyncio.ensure_future = getattr(asyncio, "async")

from .errors import ( # NOQA
    XMPPAuthError,
    XMPPCancelError,
    XMPPContinueError,
    XMPPModifyError,
    XMPPWaitError,
    ErrorCondition,
)
from .stanza import Presence, IQ, Message  # NOQA
from .structs import (  # NOQA
    JID,
    PresenceShow,
    PresenceState,
    MessageType,
    PresenceType,
    IQType,
    ErrorType,
)
from .security_layer import make as make_security_layer  # NOQA
from .node import Client, PresenceManagedClient  # NOQA

# services
from .presence import PresenceClient, PresenceServer  # NOQA
from .roster import RosterClient  # NOQA
from .disco import DiscoServer, DiscoClient  # NOQA
from .entitycaps import EntityCapsService  # NOQA
from .muc import MUCClient  # NOQA
from .pubsub import PubSubClient  # NOQA
from .shim import SHIMService  # NOQA
from .adhoc import AdHocClient, AdHocServer  # NOQA
from .avatar import AvatarService  # NOQA
from .blocking import BlockingClient  # NOQA
from .carbons import CarbonsClient  # NOQA
from .ping import PingService  # NOQA
from .pep import PEPClient  # NOQA
from .bookmarks import BookmarkClient  # NOQA
from .version import VersionServer  # NOQA
from .mdr import DeliveryReceiptsService  # NOQA

from . import httpupload


def set_strict_mode():
    from .stanza import Error
    from .stream import StanzaStream
    from . import structs
    Message.type_.type_.allow_coerce = False
    IQ.type_.type_.allow_coerce = False
    Error.type_.type_.allow_coerce = False
    Presence.type_.type_.allow_coerce = False
    StanzaStream._ALLOW_ENUM_COERCION = False
    structs._USE_COMPAT_ENUM = False



########################################################################
# File name: utils.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import asyncio
import functools


def blocking(f):
    """
    The decorated coroutine function is run using the
    :meth:`~asyncio.AbstractEventLoop.run_until_complete` method of the current
    (at the time of call) event loop.

    The decorated function behaves like a normal function and is not a
    coroutine function.

    This decorator must be applied to a coroutine function (or method).
    """

    @functools.wraps(f)
    def wrapped(*args, **kwargs):
        loop = asyncio.get_event_loop()
        return loop.run_until_complete(f(*args, **kwargs))
    return wrapped



########################################################################
# File name: service.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import asyncio

import aioxmpp
import aioxmpp.callbacks as callbacks
import aioxmpp.service as service

from aioxmpp.utils import namespaces

from . import xso as blocking_xso


class BlockingClient(service.Service):
    """
    A :class:`~aioxmpp.service.Service` implementing :xep:`Blocking
    Command <191>`.

    This service maintains the list of blocked JIDs and allows
    manipulating the blocklist.

    Attribute:

    .. autoattribute:: blocklist

    Signals:

    .. signal:: on_initial_blocklist_received(blocklist)

       Fires when the initial blocklist was received from the server.

       :param blocklist: the initial blocklist
       :type blocklist: :class:`~collections.abc.Set` of :class:`~aioxmpp.JID`

    .. signal:: on_jids_blocked(blocked_jids)

       Fires when additional JIDs are blocked.

       :param blocked_jids: the newly blocked JIDs
       :type blocked_jids: :class:`~collections.abc.Set`
           of :class:`~aioxmpp.JID`

    .. signal:: on_jids_blocked(blocked_jids)

       Fires when JIDs are unblocked.

       :param unblocked_jids: the now unblocked JIDs
       :type unblocked_jids: :class:`~collections.abc.Set`
           of :class:`~aioxmpp.JID`

    Coroutine methods:

    .. automethod:: block_jids

    .. automethod:: unblock_jids

    .. automethod:: unblock_all
    """
    ORDER_AFTER = [aioxmpp.DiscoClient]

    def __init__(self, client, **kwargs):
        super().__init__(client, **kwargs)
        self._blocklist = None
        self._lock = asyncio.Lock()
        self._disco = self.dependencies[aioxmpp.DiscoClient]

    on_jids_blocked = callbacks.Signal()
    on_jids_unblocked = callbacks.Signal()
    on_initial_blocklist_received = callbacks.Signal()

    @asyncio.coroutine
    def _check_for_blocking(self):
        server_info = yield from self._disco.query_info(
            self.client.local_jid.replace(
                resource=None,
                localpart=None,
            )
        )

        if namespaces.xep0191 not in server_info.features:
            self._blocklist = None
            raise RuntimeError("server does not support blocklists!")

    @service.depsignal(aioxmpp.Client, "before_stream_established")
    @asyncio.coroutine
    def _get_initial_blocklist(self):
        try:
            yield from self._check_for_blocking()
        except RuntimeError:
            self.logger.info(
                "server does not support block lists, skipping initial fetch"
            )
            return True

        if self._blocklist is None:
            with (yield from self._lock):
                iq = aioxmpp.IQ(
                    type_=aioxmpp.IQType.GET,
                    payload=blocking_xso.BlockList(),
                )
                result = yield from self.client.send(iq)
                self._blocklist = frozenset(result.items)
            self.on_initial_blocklist_received(self._blocklist)

        return True

    @property
    def blocklist(self):
        """
        :class:`~collections.abc.Set` of JIDs blocked by the account.
        """
        return self._blocklist

    @asyncio.coroutine
    def block_jids(self, jids_to_block):
        """
        Add the JIDs in the sequence `jids_to_block` to the client's
        blocklist.
        """
        yield from self._check_for_blocking()

        if not jids_to_block:
            return

        cmd = blocking_xso.BlockCommand(jids_to_block)
        iq = aioxmpp.IQ(
            type_=aioxmpp.IQType.SET,
            payload=cmd,
        )
        yield from self.client.send(iq)

    @asyncio.coroutine
    def unblock_jids(self, jids_to_unblock):
        """
        Remove the JIDs in the sequence `jids_to_block` from the
        client's blocklist.
        """
        yield from self._check_for_blocking()

        if not jids_to_unblock:
            return

        cmd = blocking_xso.UnblockCommand(jids_to_unblock)
        iq = aioxmpp.IQ(
            type_=aioxmpp.IQType.SET,
            payload=cmd,
        )
        yield from self.client.send(iq)

    @asyncio.coroutine
    def unblock_all(self):
        """
        Unblock all JIDs currently blocked.
        """
        yield from self._check_for_blocking()

        cmd = blocking_xso.UnblockCommand()
        iq = aioxmpp.IQ(
            type_=aioxmpp.IQType.SET,
            payload=cmd,
        )
        yield from self.client.send(iq)

    @service.iq_handler(aioxmpp.IQType.SET, blocking_xso.BlockCommand)
    @asyncio.coroutine
    def handle_block_push(self, block_command):
        diff = ()
        with (yield from self._lock):
            if self._blocklist is None:
                # this means the stream was destroyed while we were waiting for
                # the lock/while the handler was enqueued for scheduling, or
                # the server is buggy and sends pushes before we fetched the
                # blocklist
                return

            if (block_command.from_ is None or
                    block_command.from_ == self.client.local_jid.bare() or
                    # WORKAROUND: ejabberd#2287
                    block_command.from_ == self.client.local_jid):
                diff = frozenset(block_command.payload.items)
                self._blocklist |= diff
            else:
                self.logger.debug(
                    "received block push from unauthorized JID: %s",
                    block_command.from_,
                )

        if diff:
            self.on_jids_blocked(diff)

    @service.iq_handler(aioxmpp.IQType.SET, blocking_xso.UnblockCommand)
    @asyncio.coroutine
    def handle_unblock_push(self, unblock_command):
        diff = ()
        with (yield from self._lock):
            if self._blocklist is None:
                # this means the stream was destroyed while we were waiting for
                # the lock/while the handler was enqueued for scheduling, or
                # the server is buggy and sends pushes before we fetched the
                # blocklist
                return

            if (unblock_command.from_ is None or
                    unblock_command.from_ == self.client.local_jid.bare() or
                    # WORKAROUND: ejabberd#2287
                    unblock_command.from_ == self.client.local_jid):
                if not unblock_command.payload.items:
                    diff = frozenset(self._blocklist)
                    self._blocklist = frozenset()
                else:
                    diff = frozenset(unblock_command.payload.items)
                    self._blocklist -= diff
            else:
                self.logger.debug(
                    "received unblock push from unauthorized JID: %s",
                    unblock_command.from_,
                )
        if diff:
            self.on_jids_unblocked(diff)

    @service.depsignal(aioxmpp.stream.StanzaStream,
                       "on_stream_destroyed")
    def handle_stream_destroyed(self, reason):
        self._blocklist = None



########################################################################
# File name: service.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import asyncio
import contextlib
import weakref

import aioxmpp
import aioxmpp.service as service
import aioxmpp.callbacks as callbacks


class PEPClient(service.Service):
    """
    :class:`PEPClient` simplifies working with PEP services.

    .. versionchanged:: 0.10

        Before version 0.10, this service did not depend on
        :class:`aioxmpp.EntityCapsService`. This was surprising to users because
        the PEP relies on the functionality provided by the protocols
        implemented by :class:`~aioxmpp.EntityCapsService` to provide key
        features. With the release 0.10, :class:`~aioxmpp.EntityCapsService` is
        a dependency of this service. For backward compatibility, your
        application should still :meth:`aioxmpp.Client.summon` the
        :class:`~aioxmpp.EntityCapsService` explicitly.

    Compared to :class:`~aioxmpp.PubSubClient` it supports automatic
    checking for server support, a stream-lined API. It is intended to
    make PEP things easy. If you need more fine-grained control or do
    things which are not usually handled by the defaults when using PEP, use
    :class:`~aioxmpp.PubSubClient` directly.

    See :class:`register_pep_node` for the high-level interface for
    claiming a PEP node and receiving event notifications.

    There also is a low-level interface for claiming nodes:

    .. automethod:: is_claimed

    .. automethod:: claim_pep_node

    Further we have a convenience method for publishing items in the client's
    PEP service:

    .. automethod:: publish

    Use the :class:`aioxmpp.PubSubClient` For explicit subscription
    and unsubscription .
    """
    ORDER_AFTER = [
        aioxmpp.DiscoClient,
        aioxmpp.DiscoServer,
        aioxmpp.PubSubClient,
        aioxmpp.EntityCapsService,
    ]

    def __init__(self, client, **kwargs):
        super().__init__(client, **kwargs)
        self._pubsub = self.dependencies[aioxmpp.PubSubClient]
        self._disco_client = self.dependencies[aioxmpp.DiscoClient]
        self._disco_server = self.dependencies[aioxmpp.DiscoServer]

        self._pep_node_claims = weakref.WeakValueDictionary()

    def is_claimed(self, node):
        """
        Return whether `node` is claimed.
        """
        return node in self._pep_node_claims

    def claim_pep_node(self, node_namespace, *,
                       register_feature=True, notify=False):
        """
        Claim node `node_namespace`.

        :param node_namespace: the pubsub node whose events shall be
            handled.
        :param register_feature: Whether to publish the `node_namespace`
            as feature.
        :param notify: Whether to register the ``+notify`` feature to
            receive notification without explicit subscription.
        :raises RuntimeError: if a handler for `node_namespace` is already
            set.
        :returns: a :class:`~aioxmpp.pep.service.RegisteredPEPNode` instance
            representing the claim.

        .. seealso::

            :class:`aioxmpp.pep.register_pep_node`
              a descriptor which can be used with
              :class:`~aioxmpp.service.Service` subclasses to claim a PEP node
              automatically.

        This registers `node_namespace` as feature for service discovery
        unless ``register_feature=False`` is passed.

        .. note::

            For `notify` to work, it is required that
            :class:`aioxmpp.EntityCapsService` is loaded and that presence is
            re-sent soon after
            :meth:`~aioxmpp.EntityCapsService.on_ver_changed` fires. See the
            documentation of the class and the signal for details.

        """
        if node_namespace in self._pep_node_claims:
            raise RuntimeError(
                "claiming already claimed node"
            )
        registered_node = RegisteredPEPNode(
            self,
            node_namespace,
            register_feature=register_feature,
            notify=notify,
        )

        finalizer = weakref.finalize(
            registered_node,
            weakref.WeakMethod(registered_node._unregister)
        )
        # we cannot guarantee that disco is not cleared up already,
        # so we do not unclaim the feature on exit
        finalizer.atexit = False

        self._pep_node_claims[node_namespace] = registered_node

        return registered_node

    def _unclaim(self, node_namespace):
        self._pep_node_claims.pop(node_namespace)

    @asyncio.coroutine
    def available(self):
        """
        Check whether we have a PEP identity associated with our account.
        """
        disco_info = yield from self._disco_client.query_info(
            self.client.local_jid.bare()
        )

        for item in disco_info.identities.filter(attrs={"category": "pubsub"}):
            if item.type_ == "pep":
                return True
        return False

    @asyncio.coroutine
    def _check_for_pep(self):
        # XXX: should this be done when the stream connects
        # and we use the cached result later on (i.e. disable
        # the PEP service if the server does not support PEP)
        if not (yield from self.available()):
            raise RuntimeError("server does not support PEP")

    @service.depsignal(aioxmpp.PubSubClient, "on_item_published")
    def _handle_pubsub_publish(self, jid, node, item, *, message=None):
        try:
            registered_node = self._pep_node_claims[node]
        except KeyError:
            return

        # PEP requires, that notifies contain the data and that
        # the namespace of the payload corresponds to the node,
        # by enforcing this here we protect the consumers of
        # the signal.
        if (item.registered_payload is None or
                item.registered_payload.TAG[0] != node):
            self.logger.debug(
                "ignoring notify from misconfigured PEP node %s at %s",
                node, jid)
            return

        registered_node.on_item_publish(jid, node, item, message=message)

    def publish(self, node, data, *, id_=None):
        """
        Publish an item `data` in the PubSub node `node` on the
        PEP service associated with the user's JID.

        :param node: The PubSub node to publish to.
        :param data: The item to publish.
        :type data: An XSO representing the paylaod.
        :param id_: The id the published item shall have.
        :returns: The PubSub id of the published item or
            :data:`None` if it is unknown.
        :raises RuntimeError: if PEP is not supported.

        If no `id_` is given it is generated by the server (and may be
        returned).
        """
        yield from self._check_for_pep()
        return (yield from self._pubsub.publish(None, node, data, id_=id_))


class RegisteredPEPNode:
    """
    Handle for registered PEP nodes.

    *Never* instanciate this class yourself. Use
    :class:`~aioxmpp.pep.register_pep_node` or
    :attr:`~aioxmpp.pep.PEPClient.claim_pep_node` to obtain instances.

    You have to keep a reference to the instance to
    uphold the claim, when a instance is garbage
    collected it is closed automatically. It is not enough to have a
    callback registered! It is strongly recommended to explicitly
    close the registered node if it is no longer needed or to use the
    :class:`~aioxmpp.pep.register_pep_node` descriptor for automatic
    life-cycle handling.

    .. signal:: on_item_publish(jid, node, item, message=None)

       Fires when an event is received for this PEP node. The arguments
       are as for :attr:`aioxmpp.PubSubClient.on_item_publish`.

       .. warning:: Empty notifications and notifications whose
                    payload namespace does not match the node
                    namespace are filtered and will not cause
                    this signal to fire (since they do not match the
                    PEP specification).

    .. autoattribute:: notify

    .. autoattribute:: feature_registered

    .. automethod:: close

    """

    def __init__(self, pep_service, node, register_feature, notify):
        self._pep_service = pep_service
        self._node = node
        self._feature_registered = register_feature
        self._notify = notify
        self._closed = False

        if self._feature_registered:
            self._register_feature()

        if self._notify:
            self._register_notify()

    on_item_publish = callbacks.Signal()

    def _register_feature(self):
        self._pep_service._disco_server.register_feature(self._node)
        self._feature_registered = True

    def _unregister_feature(self):
        self._pep_service._disco_server.unregister_feature(self._node)
        self._feature_registered = False

    def _register_notify(self):
        self._pep_service._disco_server.register_feature(self._notify_feature)
        self._notify = True

    def _unregister_notify(self):
        self._pep_service._disco_server.unregister_feature(
            self._notify_feature)
        self._notify = False

    def _unregister(self):
        if self._notify:
            self._unregister_notify()

        if self._feature_registered:
            self._unregister_feature()

    def close(self):
        """
        Unclaim the PEP node and unregister the registered features.

        It is not necessary to call close if this claim is managed by
        :class:`~aioxmpp.pep.register_pep_node`.
        """
        if self._closed:
            return

        self._closed = True
        self._pep_service._unclaim(self.node_namespace)
        self._unregister()

    @property
    def node_namespace(self):
        """The claimed node namespace"""
        return self._node

    @property
    def _notify_feature(self):
        return self._node + "+notify"

    @property
    def notify(self):
        """
        Whether we have enabled the ``+notify`` feature to automatically
        receive notifications.

        When setting this property the feature is registered and
        unregistered appropriately.

        .. note::

            For `notify` to work, it is required that
            :class:`aioxmpp.EntityCapsService` is loaded and that presence is
            re-sent soon after
            :meth:`~aioxmpp.EntityCapsService.on_ver_changed` fires. See the
            documentation of the class and the signal for details.
        """
        return self._notify

    @notify.setter
    def notify(self, value):
        if self._closed:
            raise RuntimeError(
                "modifying a closed RegisteredPEPNode is forbidden"
            )
        # XXX: do we want to do strict type checking here?
        if bool(value) == bool(self._notify):
            return

        if self._notify:
            self._unregister_notify()
        else:
            self._register_notify()

    @property
    def feature_registered(self):
        """
        Whether we have registered the node namespace as feature.

        When setting this property the feature is registered and
        unregistered appropriately.
        """
        return self._feature_registered

    @feature_registered.setter
    def feature_registered(self, value):
        if self._closed:
            raise RuntimeError(
                "modifying a closed RegisteredPEPNode is forbidden"
            )
        # XXX: do we want to do strict type checking here?
        if bool(value) == bool(self._feature_registered):
            return

        if self._feature_registered:
            self._unregister_feature()
        else:
            self._register_feature()


class register_pep_node(service.Descriptor):
    """
    Service descriptor claiming a PEP node.

    :param node_namespace: The PubSub payload namespace to handle.
    :param register_feature: Whether to register the node namespace as feature.
    :param notify: Whether to register for notifications.

    If `notify` is :data:`True` it registers a ``+notify`` feature,
    for automatic pubsub subscription.

    The value of the descriptor on an instance is the
    :class:`~aioxmpp.pep.service.RegisteredPEPNode` object representing the
    claim.
    """

    def __init__(self, node_namespace, *, register_feature=True,
                 notify=False):
        super().__init__()
        self._node_namespace = node_namespace
        self._notify = notify
        self._register_feature = register_feature

    @property
    def node_namespace(self):
        """
        The node namespace to request notifications for.
        """
        return self._node_namespace

    @property
    def register_feature(self):
        """
        Whether we register the node namespace as feature.
        """
        return self._register_feature

    @property
    def notify(self):
        """
        Wether we register the ``+nofity`` feature.
        """
        return self._notify

    @property
    def required_dependencies(self):
        return [PEPClient]

    @contextlib.contextmanager
    def init_cm(self, instance):
        pep_client = instance.dependencies[PEPClient]
        claim = pep_client.claim_pep_node(
            self._node_namespace,
            register_feature=self._register_feature,
            notify=self._notify,
        )
        yield claim
        claim.close()

    @property
    def value_type(self):
        return RegisteredPEPNode



########################################################################
# File name: test_xso.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import contextlib
import unittest
import unittest.mock

import aioxmpp.private_xml.xso as private_xml_xso
import aioxmpp.xso as xso

from aioxmpp.utils import namespaces


@private_xml_xso.Query.as_payload_class
class FakePayload(xso.XSO):
    TAG = "tests/private_xml/test_xso.py", "payload"


class TestQuery(unittest.TestCase):

    def test_namespace(self):
        self.assertEqual(
            namespaces.xep0049,
            "jabber:iq:private"
        )

    def test_is_xso(self):
        self.assertTrue(issubclass(private_xml_xso.Query, xso.XSO))

    def test_tag(self):
        self.assertEqual(
            private_xml_xso.Query.TAG,
            (namespaces.xep0049, "query")
        )

    def test_init(self):
        payload = FakePayload()
        query = private_xml_xso.Query(payload)
        self.assertEqual(
            query.registered_payload,
            payload
        )

    def test_registered_payload_is_strict(self):
        self.assertTrue(
            private_xml_xso.Query.registered_payload.strict,
        )

    def test_as_payload_class(self):
        with contextlib.ExitStack() as stack:
            at_Query = stack.enter_context(
                unittest.mock.patch.object(
                    private_xml_xso.Query,
                    "register_child"
                )
            )

            result = private_xml_xso.Query.as_payload_class(
                unittest.mock.sentinel.cls
            )

        self.assertIs(result, unittest.mock.sentinel.cls)

        at_Query.assert_called_with(
            private_xml_xso.Query.registered_payload,
            unittest.mock.sentinel.cls
        )



########################################################################
# File name: callbacks.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
"""
:mod:`~aioxmpp.callbacks` -- Synchronous and asynchronous callbacks
###################################################################

This module provides facilities for objects to provide signals to which other
objects can connect.

Descriptor vs. ad-hoc
=====================

Descriptors can be used as class attributes and will create ad-hoc signals
dynamically for each instance. They are the most commonly used:

.. code-block:: python

   class Emitter:
       on_event = callbacks.Signal()

   def handler():
       pass

   emitter1 = Emitter()
   emitter2 = Emitter()
   emitter1.on_event.connect(handler)

   emitter1.on_event()  # calls `handler`
   emitter2.on_event()  # does not call `handler`

   # the actual signals are distinct
   assert emitter1.on_event is not emitter2.on_event

Ad-hoc signals are useful for testing and are the type of which the actual
fields are.

Signal overview
===============

.. autosummary::

   Signal
   SyncSignal
   AdHocSignal
   SyncAdHocSignal

Utilities
---------

.. autofunction:: first_signal

Signal descriptors
------------------

These descriptors can be used on classes to have attributes which are signals:

.. autoclass:: Signal

.. autoclass:: SyncSignal

Signal implementations (ad-hoc signals)
---------------------------------------

Whenever accessing an attribute using the :class:`Signal` or
:class:`SyncSignal` descriptors, an object of one of the following classes is
returned. This is where the behaviour of the signals is specified.

.. autoclass:: AdHocSignal

.. autoclass:: SyncAdHocSignal


Filters
=======

.. autoclass:: Filter

"""

import abc
import asyncio
import collections
import contextlib
import functools
import logging
import types
import weakref


logger = logging.getLogger(__name__)


def log_spawned(logger, fut):
    try:
        result = fut.result()
    except asyncio.CancelledError:
        logger.debug("spawned task was cancelled")
    except:  # NOQA
        logger.warning("spawned task raised exception", exc_info=True)
    else:
        if result is not None:
            logger.info("value returned by spawned task was ignored: %r",
                        result)


class TagListener:
    def __init__(self, ondata, onerror=None):
        self._ondata = ondata
        self._onerror = onerror

    def data(self, data):
        return self._ondata(data)

    def error(self, exc):
        if self._onerror is not None:
            return self._onerror(exc)

    def is_valid(self):
        return True


class AsyncTagListener(TagListener):
    def __init__(self, ondata, onerror=None, *, loop=None):
        super().__init__(ondata, onerror)
        self._loop = loop or asyncio.get_event_loop()

    def data(self, data):
        self._loop.call_soon(self._ondata, data)

    def error(self, exc):
        if self._onerror is not None:
            self._loop.call_soon(self._onerror, exc)


class OneshotTagListener(TagListener):
    def __init__(self, ondata, onerror=None, **kwargs):
        super().__init__(ondata, onerror=onerror, **kwargs)
        self._cancelled = False

    def data(self, data):
        super().data(data)
        return True

    def error(self, exc):
        super().error(exc)
        return True

    def cancel(self):
        self._cancelled = True

    def is_valid(self):
        return not self._cancelled and super().is_valid()


class OneshotAsyncTagListener(OneshotTagListener, AsyncTagListener):
    pass


class FutureListener:
    def __init__(self, fut):
        self.fut = fut

    def data(self, data):
        try:
            self.fut.set_result(data)
        except asyncio.futures.InvalidStateError:
            pass
        return True

    def error(self, exc):
        try:
            self.fut.set_exception(exc)
        except asyncio.futures.InvalidStateError:
            pass
        return True

    def is_valid(self):
        return not self.fut.done()


class TagDispatcher:
    def __init__(self):
        self._listeners = {}

    def add_callback(self, tag, fn):
        return self.add_listener(tag, TagListener(fn))

    def add_callback_async(self, tag, fn, *, loop=None):
        return self.add_listener(
            tag,
            AsyncTagListener(fn, loop=loop)
        )

    def add_future(self, tag, fut):
        return self.add_listener(
            tag,
            FutureListener(fut)
        )

    def add_listener(self, tag, listener):
        try:
            existing = self._listeners[tag]
            if not existing.is_valid():
                raise KeyError()
        except KeyError:
            self._listeners[tag] = listener
        else:
            raise ValueError("only one listener is allowed per tag")

    def unicast(self, tag, data):
        cb = self._listeners[tag]
        if not cb.is_valid():
            del self._listeners[tag]
            self._listeners[tag]
        if cb.data(data):
            del self._listeners[tag]

    def unicast_error(self, tag, exc):
        cb = self._listeners[tag]
        if not cb.is_valid():
            del self._listeners[tag]
            self._listeners[tag]
        if cb.error(exc):
            del self._listeners[tag]

    def remove_listener(self, tag):
        del self._listeners[tag]

    def broadcast_error(self, exc):
        for tag, listener in list(self._listeners.items()):
            if listener.is_valid() and listener.error(exc):
                del self._listeners[tag]

    def close_all(self, exc):
        self.broadcast_error(exc)
        self._listeners.clear()


class AbstractAdHocSignal:
    def __init__(self):
        super().__init__()
        self._connections = collections.OrderedDict()
        self.logger = logger

    def _connect(self, wrapper):
        token = object()
        self._connections[token] = wrapper
        return token

    def disconnect(self, token):
        """
        Disconnect the connection identified by `token`. This never raises,
        even if an invalid `token` is passed.
        """
        try:
            del self._connections[token]
        except KeyError:
            pass


class AdHocSignal(AbstractAdHocSignal):
    """
    An ad-hoc signal is a single emitter. This is where callables are connected
    to, using the :meth:`connect` method of the :class:`AdHocSignal`.

    .. automethod:: fire

    .. automethod:: connect

    .. automethod:: context_connect

    .. automethod:: future

    .. attribute:: logger

       This may be a :class:`logging.Logger` instance to allow the signal to
       log errors and debug events to a specific logger instead of the default
       logger (``aioxmpp.callbacks``).

       This attribute must not be :data:`None`, and it is initialised to the
       default logger on creation of the :class:`AdHocSignal`.

    The different ways callables can be connected to an ad-hoc signal are shown
    below:

    .. attribute:: STRONG

       Connections using this mode keep a strong reference to the callable. The
       callable is called directly, thus blocking the emission of the signal.

    .. attribute:: WEAK

       Connections using this mode keep a weak reference to the callable. The
       callable is executed directly, thus blocking the emission of the signal.

       If the weak reference is dead, it is automatically removed from the
       signals connection list. If the callable is a bound method,
       :class:`weakref.WeakMethod` is used automatically.

    For both :attr:`STRONG` and :attr:`WEAK` holds: if the callable returns a
    true value, it is disconnected from the signal.

    .. classmethod:: ASYNC_WITH_LOOP(loop)

       This mode requires an :mod:`asyncio` event loop as argument. When the
       signal is emitted, the callable is not called directly. Instead, it is
       enqueued for calling with the event loop using
       :meth:`asyncio.BaseEventLoop.call_soon`. If :data:`None` is passed as
       `loop`, the loop is obtained from :func:`asyncio.get_event_loop` at
       connect time.

       A strong reference is held to the callable.

       Connections using this mode are never removed automatically from the
       signals connection list. You have to use :meth:`disconnect` explicitly.

    .. attribute:: AUTO_FUTURE

       Instead of a callable, a :class:`asyncio.Future` must be passed when
       using this mode.

       This mode can only be used for signals which send at most one
       argument. If no argument is sent, the :meth:`~asyncio.Future.set_result`
       method is called with :data:`None`.

       If one argument is sent and it is an instance of :class:`Exception`, it
       is passed to :meth:`~asyncio.Future.set_exception`. Otherwise, if one
       argument is sent, it is passed to
       :meth:`~asyncio.Future.set_exception`.

       In any case, the future is removed after the next emission of the
       signal.

    .. classmethod:: SPAWN_WITH_LOOP(loop)

       This mode requires an :mod:`asyncio` event loop as argument and a
       coroutine to be passed to :meth:`connect`. If :data:`None` is passed as
       `loop`, the loop is obtained from :func:`asyncio.get_event_loop` at
       connect time.

       When the signal is emitted, the coroutine is spawned using
       :func:`asyncio.ensure_future` in the given `loop`, with the arguments
       passed to the signal.

       A strong reference is held to the coroutine.

       Connections using this mode are never removed automatically from the
       signals connection list. You have to use :meth:`disconnect` explicitly.

       If the spawned coroutine returns with an exception or a non-:data:`None`
       return value, a message is logged, with the following log levels:

       * Return with non-:data:`None` value: :data:`logging.INFO`
       * Raises :class:`asyncio.CancelledError`: :data:`logging.DEBUG`
       * Raises any other exception: :data:`logging.WARNING`

       .. versionadded:: 0.6

    .. automethod:: disconnect

    """

    @classmethod
    def STRONG(cls, f):
        if not hasattr(f, "__call__"):
            raise TypeError("must be callable, got {!r}".format(f))
        return functools.partial(cls._strong_wrapper, f)

    @classmethod
    def ASYNC_WITH_LOOP(cls, loop):
        if loop is None:
            loop = asyncio.get_event_loop()

        def create_wrapper(f):
            if not hasattr(f, "__call__"):
                raise TypeError("must be callable, got {!r}".format(f))
            return functools.partial(cls._async_wrapper,
                                     f,
                                     loop)

        return create_wrapper

    @classmethod
    def WEAK(cls, f):
        if not hasattr(f, "__call__"):
            raise TypeError("must be callable, got {!r}".format(f))
        if isinstance(f, types.MethodType):
            ref = weakref.WeakMethod(f)
        else:
            ref = weakref.ref(f)
        return functools.partial(cls._weakref_wrapper, ref)

    @classmethod
    def AUTO_FUTURE(cls, f):
        def future_wrapper(args, kwargs):
            if kwargs:
                raise TypeError("keyword arguments not supported")
            if len(args) > 0:
                try:
                    arg, = args
                except ValueError:
                    raise TypeError("too many arguments") from None
            else:
                arg = None
            if f.done():
                return
            if isinstance(arg, Exception):
                f.set_exception(arg)
            else:
                f.set_result(arg)
        return future_wrapper

    @classmethod
    def SPAWN_WITH_LOOP(cls, loop):
        loop = asyncio.get_event_loop() if loop is None else loop

        def spawn(f):
            if not asyncio.iscoroutinefunction(f):
                raise TypeError("must be coroutine, got {!r}".format(f))

            def wrapper(args, kwargs):
                task = asyncio.ensure_future(f(*args, **kwargs), loop=loop)
                task.add_done_callback(
                    functools.partial(
                        log_spawned,
                        logger,
                    )
                )
                return True

            return wrapper

        return spawn

    @staticmethod
    def _async_wrapper(f, loop, args, kwargs):
        if kwargs:
            functools.partial(f, *args, **kwargs)
        loop.call_soon(f, *args)
        return True

    @staticmethod
    def _weakref_wrapper(fref, args, kwargs):
        f = fref()
        if f is None:
            return False
        return not f(*args, **kwargs)

    @staticmethod
    def _strong_wrapper(f, args, kwargs):
        return not f(*args, **kwargs)

    def connect(self, f, mode=None):
        """
        Connect an object `f` to the signal. The type the object needs to have
        depends on `mode`, but usually it needs to be a callable.

        :meth:`connect` returns an opaque token which can be used with
        :meth:`disconnect` to disconnect the object from the signal.

        The default value for `mode` is :attr:`STRONG`. Any decorator can be
        used as argument for `mode` and it is applied to `f`. The result is
        stored internally and is what will be called when the signal is being
        emitted.

        If the result of `mode` returns a false value during emission, the
        connection is removed.

        .. note::

           The return values required by the callable returned by `mode` and
           the one required by a callable passed to `f` using the predefined
           modes are complementary!

           A callable `f` needs to return true to be removed from the
           connections, while a callable returned by the `mode` decorator needs
           to return false.

        Existing modes are listed below.
        """

        mode = mode or self.STRONG
        self.logger.debug("connecting %r with mode %r", f, mode)
        return self._connect(mode(f))

    def context_connect(self, f, mode=None):
        """
        This returns a *context manager*. When entering the context, `f` is
        connected to the :class:`AdHocSignal` using `mode`. When leaving the
        context (no matter whether with or without exception), the connection
        is disconnected.

        .. seealso::

           The returned object is an instance of
           :class:`SignalConnectionContext`.

        """
        return SignalConnectionContext(self, f, mode=mode)

    def fire(self, *args, **kwargs):
        """
        Emit the signal, calling all connected objects in-line with the given
        arguments and in the order they were registered.

        :class:`AdHocSignal` provides full isolation with respect to
        exceptions. If a connected listener raises an exception, the other
        listeners are executed as normal, but the raising listener is removed
        from the signal. The exception is logged to :attr:`logger` and *not*
        re-raised, so that the caller of the signal is also not affected.

        Instead of calling :meth:`fire` explicitly, the ad-hoc signal object
        itself can be called, too.
        """
        for token, wrapper in list(self._connections.items()):
            try:
                keep = wrapper(args, kwargs)
            except Exception:
                self.logger.exception("listener attached to signal raised")
                keep = False
            if not keep:
                del self._connections[token]

    def future(self):
        """
        Return a :class:`asyncio.Future` which has been :meth:`connect`\ -ed
        using :attr:`AUTO_FUTURE`.

        The token returned by :meth:`connect` is not returned; to remove the
        future from the signal, just cancel it.
        """
        fut = asyncio.Future()
        self.connect(fut, self.AUTO_FUTURE)
        return fut

    __call__ = fire


class SyncAdHocSignal(AbstractAdHocSignal):
    """
    A synchronous ad-hoc signal is like :class:`AdHocSignal`, but for
    coroutines instead of ordinary callables.

    .. automethod:: connect

    .. automethod:: context_connect

    .. automethod:: fire

    .. automethod:: disconnect
    """

    def connect(self, coro):
        """
        The coroutine `coro` is connected to the signal. The coroutine must
        return a true value, unless it wants to be disconnected from the
        signal.

        .. note::

           This is different from the return value convention with
           :attr:`AdHocSignal.STRONG` and :attr:`AdHocSignal.WEAK`.

        :meth:`connect` returns a token which can be used with
        :meth:`disconnect` to disconnect the coroutine.
        """
        self.logger.debug("connecting %r", coro)
        return self._connect(coro)

    def context_connect(self, coro):
        """
        This returns a *context manager*. When entering the context, `coro` is
        connected to the :class:`SyncAdHocSignal`. When leaving the context (no
        matter whether with or without exception), the connection is
        disconnected.

        .. seealso::

           The returned object is an instance of
           :class:`SignalConnectionContext`.

        """
        return SignalConnectionContext(self, coro)

    @asyncio.coroutine
    def fire(self, *args, **kwargs):
        """
        Emit the signal, calling all coroutines in-line with the given
        arguments and in the order they were registered.

        This is obviously a coroutine.

        Instead of calling :meth:`fire` explicitly, the ad-hoc signal object
        itself can be called, too.
        """
        for token, coro in list(self._connections.items()):
            keep = yield from coro(*args, **kwargs)
            if not keep:
                del self._connections[token]

    __call__ = fire


class SignalConnectionContext:
    def __init__(self, signal, *args, **kwargs):
        self._signal = signal
        self._args = args
        self._kwargs = kwargs

    def __enter__(self):
        try:
            token = self._signal.connect(*self._args, **self._kwargs)
        finally:
            del self._args
            del self._kwargs
        self._token = token
        return token

    def __exit__(self, exc_type, exc_value, traceback):
        self._signal.disconnect(self._token)
        return False


class AbstractSignal(metaclass=abc.ABCMeta):
    def __init__(self, *, doc=None):
        super().__init__()
        self.__doc__ = doc
        self._instances = weakref.WeakKeyDictionary()

    @abc.abstractclassmethod
    def make_adhoc_signal(cls):
        pass

    def __get__(self, instance, owner):
        if instance is None:
            return self
        try:
            return self._instances[instance]
        except KeyError:
            new = self.make_adhoc_signal()
            self._instances[instance] = new
            return new

    def __set__(self, instance, value):
        raise AttributeError("cannot override Signal attribute")

    def __delete__(self, instance):
        raise AttributeError("cannot override Signal attribute")


class Signal(AbstractSignal):
    """
    A descriptor which returns per-instance :class:`AdHocSignal` objects on
    attribute access.

    Example use:

    .. code-block:: python

       class Foo:
           on_event = Signal()

       f = Foo()
       assert isinstance(f.on_event, AdHocSignal)
       assert f.on_event is f.on_event
       assert Foo().on_event is not f.on_event

    """

    @classmethod
    def make_adhoc_signal(cls):
        return AdHocSignal()


class SyncSignal(AbstractSignal):
    """
    A descriptor which returns per-instance :class:`SyncAdHocSignal` objects on
    attribute access.

    Example use:

    .. code-block:: python

       class Foo:
           on_event = SyncSignal()

       f = Foo()
       assert isinstance(f.on_event, SyncAdHocSignal)
       assert f.on_event is f.on_event
       assert Foo().on_event is not f.on_event
    """

    @classmethod
    def make_adhoc_signal(cls):
        return SyncAdHocSignal()


class Filter:
    """
    A filter chain for arbitrary data.

    This is used for example in :class:`~.stream.StanzaStream` to allow
    services and applications to filter inbound and outbound stanzas.

    Each function registered with the filter receives at least one argument.
    This argument is the object which is to be filtered. The function must
    return the object, a replacement or :data:`None`. If :data:`None` is
    returned, the filter chain aborts and further functions are not called.
    Otherwise, the next function is called with the result of the previous
    function until the filter chain is complete.

    Other arguments passed to :meth:`filter` are passed unmodified to each
    function called; only the first argument is subject to filtering.

    .. versionchanged:: 0.9

       This class was formerly available at :class:`aioxmpp.stream.Filter`.

    .. automethod:: register

    .. automethod:: filter

    .. automethod:: unregister

    .. automethod:: context_register(func[, order])
    """

    class Token:
        def __str__(self):
            return "<{}.{} 0x{:x}>".format(
                type(self).__module__,
                type(self).__qualname__,
                id(self))

    def __init__(self):
        super().__init__()
        self._filter_order = []

    def register(self, func, order):
        """
        Add a function to the filter chain.

        :param func: A callable which is to be added to the filter chain.
        :param order: An object indicating the ordering of the function
                      relative to the others.
        :return: Token representing the registration.

        Register the function `func` as a filter into the chain. `order` must
        be a value which is used as a sorting key to order the functions
        registered in the chain.

        The type of `order` depends on the use of the filter, as does the
        number of arguments and keyword arguments which `func` must accept.
        This will generally be documented at the place where the
        :class:`Filter` is used.

        Functions with the same order are sorted in the order of their
        addition, with the function which was added earliest first.

        Remember that all values passed to `order` which are registered at the
        same time in the same :class:`Filter` need to be totally orderable with
        respect to each other.

        The returned token can be used to :meth:`unregister` a filter.
        """
        token = self.Token()
        self._filter_order.append((order, token, func))
        self._filter_order.sort(key=lambda x: x[0])
        return token

    def filter(self, obj, *args, **kwargs):
        """
        Filter the given object through the filter chain.

        :param obj: The object to filter
        :param args: Additional arguments to pass to each filter function.
        :param kwargs: Additional keyword arguments to pass to each filter
                       function.
        :return: The filtered object or :data:`None`

        See the documentation of :class:`Filter` on how filtering operates.

        Returns the object returned by the last function in the filter chain or
        :data:`None` if any function returned :data:`None`.
        """
        for _, _, func in self._filter_order:
            obj = func(obj, *args, **kwargs)
            if obj is None:
                return None
        return obj

    def unregister(self, token_to_remove):
        """
        Unregister a filter function.

        :param token_to_remove: The token as returned by :meth:`register`.

        Unregister a function from the filter chain using the token returned by
        :meth:`register`.
        """
        for i, (_, token, _) in enumerate(self._filter_order):
            if token == token_to_remove:
                break
        else:
            raise ValueError("unregistered token: {!r}".format(
                token_to_remove))
        del self._filter_order[i]

    @contextlib.contextmanager
    def context_register(self, func, *args):
        """
        :term:`Context manager <context manager>` which temporarily registers a
        filter function.

        :param func: The filter function to register.
        :param order: The sorting key for the filter function.
        :rtype: :term:`context manager`
        :return: Context manager which temporarily registers the filter
                 function.

        If :meth:`register` does not require `order` because it has been
        overridden in a subclass, the `order` argument can be omitted here,
        too.

        .. versionadded:: 0.9
        """
        token = self.register(func, *args)
        try:
            yield
        finally:
            self.unregister(token)


def first_signal(*signals):
    """
    Connect to multiple signals and wait for the first to emit.

    :param signals: Signals to connect to.
    :type signals: :class:`AdHocSignal`
    :return: An awaitable for the first signal to emit.

    The awaitable returns the first argument passed to the signal. If the first
    argument is an exception, the exception is re-raised from the awaitable.

    A common use-case is a situation where a class exposes a "on_finished" type
    signal and an "on_failure" type signal. :func:`first_signal` can be used
    to combine those nicely::

        # e.g. a aioxmpp.im.conversation.AbstractConversation
        conversation = ...
        await first_signal(
            # emits without arguments when the conversation is successfully
            # entered
            conversation.on_enter,
            # emits with an exception when entering the conversation fails
            conversation.on_failure,
        )
        # await first_signal(...) will either raise an exception (failed) or
        # return None (success)

    .. warning::

        Only works with signals which emit with zero or one argument. Signals
        which emit with more than one argument or with keyword arguments are
        silently ignored! (Thus, if only such signals are connected, the
        future will never complete.)

        (This is a side-effect of the implementation of
        :meth:`AdHocSignal.AUTO_FUTURE`).

    .. note::

        Does not work with coroutine signals (:class:`SyncAdHocSignal`).
    """

    fut = asyncio.Future()
    for signal in signals:
        signal.connect(fut, signal.AUTO_FUTURE)
    return fut



########################################################################
# File name: __init__.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
"""
:mod:`~aioxmpp.disco` --- Service discovery support (:xep:`0030`)
#################################################################

This module provides support for :xep:`Service Discovery <30>`. For this, it
provides a :class:`~aioxmpp.service.Service` subclass which can be loaded into
a client using :meth:`.Client.summon`.

Services
========

The following services are provided by this subpackage and available directly
from :mod:`aioxmpp`:

.. currentmodule:: aioxmpp

.. autosummary::
   :nosignatures:

   DiscoServer
   DiscoClient

.. versionchanged:: 0.8

   Prior to version 0.8, both services were provided by a single class
   (:class:`aioxmpp.disco.Service`). This is not the case anymore, and there is
   no replacement.

   If you need to write backwards compatible code, you could be doing something
   like this::

     try:
         aioxmpp.DiscoServer
     except AttributeError:
         aioxmpp.DiscoServer = aioxmpp.disco.Service
         aioxmpp.DiscoClient = aioxmpp.disco.Service

   This should work, because the old :class:`Service` class provided the
   features of both of the individual classes.

The detailed documentation of the classes follows:

.. autoclass:: DiscoServer

.. autoclass:: DiscoClient

.. currentmodule:: aioxmpp.disco

Entity information
------------------

.. autoclass:: Node

.. autoclass:: StaticNode

.. autoclass:: mount_as_node

.. autoclass:: register_feature

.. autoclass:: RegisteredFeature

.. module:: aioxmpp.disco.xso

.. currentmodule:: aioxmpp.disco.xso

:mod:`.disco.xso` --- IQ payloads
=================================

The submodule :mod:`aioxmpp.disco.xso` contains the :class:`~aioxmpp.xso.XSO`
classes which describe the IQ payloads used by this subpackage.

You will encounter some of these in return values, but there should never be a
need to construct them by yourself; the :class:`~aioxmpp.disco.Service` handles
it all.

Information queries
-------------------

.. autoclass:: InfoQuery(*[, identities][, features][, node])

.. autoclass:: Feature(*[, var])

.. autoclass:: Identity(*[, category][, type_][, name][, lang])

Item queries
------------

.. autoclass:: ItemsQuery(*[, node][, items])

.. autoclass:: Item(*[, jid][, name][, node])

.. currentmodule:: aioxmpp.disco

"""

from . import xso  # NOQA
from .service import (DiscoClient, DiscoServer, Node, StaticNode,  # NOQA
                      mount_as_node, register_feature, RegisteredFeature)



########################################################################
# File name: test_e2e.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import logging
import unittest
import uuid

import aioxmpp
import aioxmpp.pubsub
import aioxmpp.xso

from aioxmpp.utils import namespaces

from aioxmpp.e2etest import (
    TestCase,
    blocking,
    blocking_timed,
    require_feature_subset,
    require_pep,
)


@aioxmpp.pubsub.xso.as_payload_class
class Payload(aioxmpp.xso.XSO):
    TAG = (namespaces.aioxmpp_test, "foo")


class TestOwnerUseCases(TestCase):
    @require_feature_subset(
        {
            namespaces.xep0060_features.PURGE_NODES.value,
            namespaces.xep0060_features.MODIFY_AFFILIATIONS.value,
            namespaces.xep0060_features.PUBLISH.value,
            namespaces.xep0060_features.PUBLISHER_AFFILIATION.value,
        },
        required_subset={
            namespaces.xep0060_features.CREATE_NODES.value,
            namespaces.xep0060_features.DELETE_NODES.value,
        }
    )
    @blocking
    def setUp(self, pubsub_jid, supported_features):
        logging.info("using pubsub service %s, which provides %r",
                     pubsub_jid, supported_features)

        self.peer = pubsub_jid
        self.peer_features = supported_features

        self.owner = yield from self.provisioner.get_connected_client()
        self.pubsub = self.owner.summon(aioxmpp.PubSubClient)

        self.other = yield from self.provisioner.get_connected_client()

        self.node = str(uuid.uuid4())

        yield from self.pubsub.create(self.peer, self.node)

    def _require_features(self, features):
        features = set(features)
        if features & self.peer_features != features:
            raise unittest.SkipTest(
                "selected PubSub provider ({!r}) does not provide {!r}".format(
                    self.peer,
                    features,
                )
            )

    @blocking_timed
    def test_publish_and_purge(self):
        self._require_features({
            namespaces.xep0060_features.PURGE_NODES.value
        })

        payload = Payload()

        id_ = yield from self.pubsub.publish(
            self.peer,
            self.node,
            payload,
        )
        items = yield from self.pubsub.get_items(
            self.peer,
            self.node,
        )

        self.assertEqual(len(items.payload.items), 1)
        yield from self.pubsub.purge(self.peer, self.node)

        # using get items instead of get_items_by_id to work around ejabberd
        # #2288
        items = yield from self.pubsub.get_items(
            self.peer,
            self.node,
        )
        self.assertEqual(len(items.payload.items), 0)

    @blocking
    def tearDown(self):
        yield from self.pubsub.delete(self.peer, self.node)



########################################################################
# File name: test_caps390.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import contextlib
import io
import pathlib
import unittest
import unittest.mock
import urllib.parse

import aioxmpp
import aioxmpp.entitycaps.caps390 as caps390
import aioxmpp.entitycaps.xso as caps_xso


def _parse_testcase(data):
    src = io.BytesIO(data)
    return aioxmpp.xml.read_single_xso(src, aioxmpp.disco.xso.InfoQuery)


TEST_SMALL = _parse_testcase(b"""\
<query xmlns="http://jabber.org/protocol/disco#info">
  <identity category="client" name="BombusMod" type="mobile"/>
  <feature var="http://jabber.org/protocol/si"/>
  <feature var="http://jabber.org/protocol/bytestreams"/>
  <feature var="http://jabber.org/protocol/chatstates"/>
  <feature var="http://jabber.org/protocol/disco#info"/>
  <feature var="http://jabber.org/protocol/disco#items"/>
  <feature var="urn:xmpp:ping"/>
  <feature var="jabber:iq:time"/>
  <feature var="jabber:iq:privacy"/>
  <feature var="jabber:iq:version"/>
  <feature var="http://jabber.org/protocol/rosterx"/>
  <feature var="urn:xmpp:time"/>
  <feature var="jabber:x:oob"/>
  <feature var="http://jabber.org/protocol/ibb"/>
  <feature var="http://jabber.org/protocol/si/profile/file-transfer"/>
  <feature var="urn:xmpp:receipts"/>
  <feature var="jabber:iq:roster"/>
  <feature var="jabber:iq:last"/>
</query>""")

TEST_LARGE = _parse_testcase("""\
<query xmlns="http://jabber.org/protocol/disco#info">
  <identity category="client" name="Tkabber" type="pc" xml:lang="en"/>
  <identity category="client" name="Ткаббер" type="pc" xml:lang="ru"/>
  <feature var="games:board"/>
  <feature var="http://jabber.org/protocol/activity"/>
  <feature var="http://jabber.org/protocol/activity+notify"/>
  <feature var="http://jabber.org/protocol/bytestreams"/>
  <feature var="http://jabber.org/protocol/chatstates"/>
  <feature var="http://jabber.org/protocol/commands"/>
  <feature var="http://jabber.org/protocol/disco#info"/>
  <feature var="http://jabber.org/protocol/disco#items"/>
  <feature var="http://jabber.org/protocol/evil"/>
  <feature var="http://jabber.org/protocol/feature-neg"/>
  <feature var="http://jabber.org/protocol/geoloc"/>
  <feature var="http://jabber.org/protocol/geoloc+notify"/>
  <feature var="http://jabber.org/protocol/ibb"/>
  <feature var="http://jabber.org/protocol/iqibb"/>
  <feature var="http://jabber.org/protocol/mood"/>
  <feature var="http://jabber.org/protocol/mood+notify"/>
  <feature var="http://jabber.org/protocol/rosterx"/>
  <feature var="http://jabber.org/protocol/si"/>
  <feature var="http://jabber.org/protocol/si/profile/file-transfer"/>
  <feature var="http://jabber.org/protocol/tune"/>
  <feature var="http://www.facebook.com/xmpp/messages"/>
  <feature var="http://www.xmpp.org/extensions/xep-0084.html#ns-metadata+notify"/>
  <feature var="jabber:iq:avatar"/>
  <feature var="jabber:iq:browse"/>
  <feature var="jabber:iq:dtcp"/>
  <feature var="jabber:iq:filexfer"/>
  <feature var="jabber:iq:ibb"/>
  <feature var="jabber:iq:inband"/>
  <feature var="jabber:iq:jidlink"/>
  <feature var="jabber:iq:last"/>
  <feature var="jabber:iq:oob"/>
  <feature var="jabber:iq:privacy"/>
  <feature var="jabber:iq:roster"/>
  <feature var="jabber:iq:time"/>
  <feature var="jabber:iq:version"/>
  <feature var="jabber:x:data"/>
  <feature var="jabber:x:event"/>
  <feature var="jabber:x:oob"/>
  <feature var="urn:xmpp:avatar:metadata+notify"/>
  <feature var="urn:xmpp:ping"/>
  <feature var="urn:xmpp:receipts"/>
  <feature var="urn:xmpp:time"/>
  <x xmlns="jabber:x:data" type="result">
    <field type="hidden" var="FORM_TYPE">
      <value>urn:xmpp:dataforms:softwareinfo</value>
    </field>
    <field var="software">
      <value>Tkabber</value>
    </field>
    <field var="software_version">
      <value>0.11.1-svn-20111216-mod (Tcl/Tk 8.6b2)</value>
    </field>
    <field var="os">
      <value>Windows</value>
    </field>
    <field var="os_version">
      <value>XP</value>
    </field>
  </x>
</query>""".encode("utf-8"))


class Test_process_features(unittest.TestCase):
    def test_on_small_testcase(self):
        self.assertEqual(
            caps390._process_features(TEST_SMALL.features),
            b'http://jabber.org/protocol/bytestreams\x1fhttp://jabber.org/prot'
            b'ocol/chatstates\x1fhttp://jabber.org/protocol/disco#info\x1fhttp'
            b'://jabber.org/protocol/disco#items\x1fhttp://jabber.org/protocol'
            b'/ibb\x1fhttp://jabber.org/protocol/rosterx\x1fhttp://jabber.org/'
            b'protocol/si\x1fhttp://jabber.org/protocol/si/profile/file-transf'
            b'er\x1fjabber:iq:last\x1fjabber:iq:privacy\x1fjabber:iq:roster'
            b'\x1fjabber:iq:time\x1fjabber:iq:version\x1fjabber:x:oob\x1furn:x'
            b'mpp:ping\x1furn:xmpp:receipts\x1furn:xmpp:time\x1f\x1c'
        )

    def test_on_large_testcase(self):
        self.assertEqual(
            caps390._process_features(TEST_LARGE.features),
            b'games:board\x1fhttp://jabber.org/protocol/activity\x1fhttp://jab'
            b'ber.org/protocol/activity+notify\x1fhttp://jabber.org/protocol/b'
            b'ytestreams\x1fhttp://jabber.org/protocol/chatstates\x1fhttp://ja'
            b'bber.org/protocol/commands\x1fhttp://jabber.org/protocol/disco#i'
            b'nfo\x1fhttp://jabber.org/protocol/disco#items\x1fhttp://jabber.o'
            b'rg/protocol/evil\x1fhttp://jabber.org/protocol/feature-neg\x1fht'
            b'tp://jabber.org/protocol/geoloc\x1fhttp://jabber.org/protocol/ge'
            b'oloc+notify\x1fhttp://jabber.org/protocol/ibb\x1fhttp://jabber.o'
            b'rg/protocol/iqibb\x1fhttp://jabber.org/protocol/mood\x1fhttp://j'
            b'abber.org/protocol/mood+notify\x1fhttp://jabber.org/protocol/ros'
            b'terx\x1fhttp://jabber.org/protocol/si\x1fhttp://jabber.org/proto'
            b'col/si/profile/file-transfer\x1fhttp://jabber.org/protocol/tune'
            b'\x1fhttp://www.facebook.com/xmpp/messages\x1fhttp://www.xmpp.org'
            b'/extensions/xep-0084.html#ns-metadata+notify\x1fjabber:iq:avatar'
            b'\x1fjabber:iq:browse\x1fjabber:iq:dtcp\x1fjabber:iq:filexfer\x1f'
            b'jabber:iq:ibb\x1fjabber:iq:inband\x1fjabber:iq:jidlink\x1fjabber'
            b':iq:last\x1fjabber:iq:oob\x1fjabber:iq:privacy\x1fjabber:iq:rost'
            b'er\x1fjabber:iq:time\x1fjabber:iq:version\x1fjabber:x:data\x1fja'
            b'bber:x:event\x1fjabber:x:oob\x1furn:xmpp:avatar:metadata+notify'
            b'\x1furn:xmpp:ping\x1furn:xmpp:receipts\x1furn:xmpp:time\x1f\x1c'
        )


class Test_process_identities(unittest.TestCase):
    def test_on_small_testcase(self):
        self.assertEqual(
            caps390._process_identities(TEST_SMALL.identities),
            b'client\x1fmobile\x1f\x1fBombusMod\x1f\x1e\x1c'
        )

    def test_on_large_testcase(self):
        self.assertEqual(
            caps390._process_identities(TEST_LARGE.identities),
            b'client\x1fpc\x1fen\x1fTkabber\x1f\x1eclient\x1fpc\x1fru\x1f\xd0'
            b'\xa2\xd0\xba\xd0\xb0\xd0\xb1\xd0\xb1\xd0\xb5\xd1\x80\x1f\x1e\x1c'
        )


class Test_process_extensions(unittest.TestCase):
    def test_on_small_testcase(self):
        self.assertEqual(
            caps390._process_extensions(TEST_SMALL.exts),
            b'\x1c'
        )

    def test_on_large_testcase(self):
        self.assertEqual(
            caps390._process_extensions(TEST_LARGE.exts),
            b'FORM_TYPE\x1furn:xmpp:dataforms:softwareinfo\x1f\x1eos\x1fWindow'
            b's\x1f\x1eos_version\x1fXP\x1f\x1esoftware\x1fTkabber\x1f\x1esoft'
            b'ware_version\x1f0.11.1-svn-20111216-mod (Tcl/Tk 8.6b2)\x1f\x1e'
            b'\x1d\x1c'
        )


class Test_get_hash_input(unittest.TestCase):
    def test_on_small_testcase(self):
        self.assertEqual(
            caps390._get_hash_input(TEST_SMALL),
            b'http://jabber.org/protocol/bytestreams\x1fhttp://jabber.org/prot'
            b'ocol/chatstates\x1fhttp://jabber.org/protocol/disco#info\x1fhttp'
            b'://jabber.org/protocol/disco#items\x1fhttp://jabber.org/protocol'
            b'/ibb\x1fhttp://jabber.org/protocol/rosterx\x1fhttp://jabber.org/'
            b'protocol/si\x1fhttp://jabber.org/protocol/si/profile/file-transf'
            b'er\x1fjabber:iq:last\x1fjabber:iq:privacy\x1fjabber:iq:roster'
            b'\x1fjabber:iq:time\x1fjabber:iq:version\x1fjabber:x:oob\x1furn:x'
            b'mpp:ping\x1furn:xmpp:receipts\x1furn:xmpp:time\x1f\x1cclient\x1f'
            b'mobile\x1f\x1fBombusMod\x1f\x1e\x1c\x1c'
        )

    def test_on_large_testcase(self):
        self.assertEqual(
            caps390._get_hash_input(TEST_LARGE),
            b'games:board\x1fhttp://jabber.org/protocol/activity\x1fhttp://jab'
            b'ber.org/protocol/activity+notify\x1fhttp://jabber.org/protocol/b'
            b'ytestreams\x1fhttp://jabber.org/protocol/chatstates\x1fhttp://ja'
            b'bber.org/protocol/commands\x1fhttp://jabber.org/protocol/disco#i'
            b'nfo\x1fhttp://jabber.org/protocol/disco#items\x1fhttp://jabber.o'
            b'rg/protocol/evil\x1fhttp://jabber.org/protocol/feature-neg\x1fht'
            b'tp://jabber.org/protocol/geoloc\x1fhttp://jabber.org/protocol/ge'
            b'oloc+notify\x1fhttp://jabber.org/protocol/ibb\x1fhttp://jabber.o'
            b'rg/protocol/iqibb\x1fhttp://jabber.org/protocol/mood\x1fhttp://j'
            b'abber.org/protocol/mood+notify\x1fhttp://jabber.org/protocol/ros'
            b'terx\x1fhttp://jabber.org/protocol/si\x1fhttp://jabber.org/proto'
            b'col/si/profile/file-transfer\x1fhttp://jabber.org/protocol/tune'
            b'\x1fhttp://www.facebook.com/xmpp/messages\x1fhttp://www.xmpp.org'
            b'/extensions/xep-0084.html#ns-metadata+notify\x1fjabber:iq:avatar'
            b'\x1fjabber:iq:browse\x1fjabber:iq:dtcp\x1fjabber:iq:filexfer\x1f'
            b'jabber:iq:ibb\x1fjabber:iq:inband\x1fjabber:iq:jidlink\x1fjabber'
            b':iq:last\x1fjabber:iq:oob\x1fjabber:iq:privacy\x1fjabber:iq:rost'
            b'er\x1fjabber:iq:time\x1fjabber:iq:version\x1fjabber:x:data\x1fja'
            b'bber:x:event\x1fjabber:x:oob\x1furn:xmpp:avatar:metadata+notify'
            b'\x1furn:xmpp:ping\x1furn:xmpp:receipts\x1furn:xmpp:time\x1f\x1cc'
            b'lient\x1fpc\x1fen\x1fTkabber\x1f\x1eclient\x1fpc\x1fru\x1f\xd0'
            b'\xa2\xd0\xba\xd0\xb0\xd0\xb1\xd0\xb1\xd0\xb5\xd1\x80\x1f\x1e\x1c'
            b'FORM_TYPE\x1furn:xmpp:dataforms:softwareinfo\x1f\x1eos\x1fWindow'
            b's\x1f\x1eos_version\x1fXP\x1f\x1esoftware\x1fTkabber\x1f\x1esoft'
            b'ware_version\x1f0.11.1-svn-20111216-mod (Tcl/Tk 8.6b2)\x1f\x1e'
            b'\x1d\x1c'
        )


class Test_calculate_hash(unittest.TestCase):
    def test_uses_and_hash_from_algo(self):
        with contextlib.ExitStack() as stack:
            hash_from_algo = stack.enter_context(
                unittest.mock.patch("aioxmpp.hashes.hash_from_algo")
            )

            _get_hash_input = stack.enter_context(
                unittest.mock.patch(
                    "aioxmpp.entitycaps.caps390._get_hash_input"
                )
            )

            result = caps390._calculate_hash(
                unittest.mock.sentinel.algo,
                unittest.mock.sentinel.hash_input,
            )

        hash_from_algo.assert_called_once_with(unittest.mock.sentinel.algo)
        _get_hash_input.assert_not_called()

        hash_from_algo().update.assert_called_once_with(
            unittest.mock.sentinel.hash_input
        )
        hash_from_algo().digest.assert_called_once_with()

        self.assertEqual(
            result,
            hash_from_algo().digest(),
        )


class TestKey(unittest.TestCase):
    def test_init_default(self):
        with self.assertRaises(TypeError):
            caps390.Key()

    def test_init_args(self):
        k = caps390.Key("algo", b"digest")

        self.assertEqual(
            k.algo,
            "algo",
        )
        self.assertEqual(
            k.digest,
            b"digest",
        )

    def test_init_kwargs(self):
        k = caps390.Key(algo="somealgo", digest=b"somedigest")

        self.assertEqual(
            k.algo,
            "somealgo",
        )
        self.assertEqual(
            k.digest,
            b"somedigest",
        )

    def test_hashable(self):
        k1 = caps390.Key("algo", b"digest")
        k2 = caps390.Key("algo", b"digest")

        self.assertEqual(hash(k1), hash(k2))

    def test_equality(self):
        k1 = caps390.Key("algo", b"digest")
        k2 = caps390.Key("algo", b"digest")
        k3 = caps390.Key("somealgo", b"otherdigest")

        self.assertTrue(k1 == k2)
        self.assertFalse(k1 != k2)
        self.assertFalse(k1 == k3)
        self.assertTrue(k1 != k3)

        self.assertTrue(k2 == k1)
        self.assertFalse(k2 != k1)
        self.assertFalse(k2 == k3)
        self.assertTrue(k2 != k3)

        self.assertFalse(k3 == k1)
        self.assertTrue(k3 != k1)
        self.assertFalse(k3 == k2)
        self.assertTrue(k3 != k2)

    def test_node(self):
        k1 = caps390.Key("algo", b"digest")

        self.assertEqual(
            k1.node,
            "urn:xmpp:caps#algo.ZGlnZXN0"
        )

        k1 = caps390.Key("somealgo", b"otherdigest")

        self.assertEqual(
            k1.node,
            "urn:xmpp:caps#somealgo.b3RoZXJkaWdlc3Q="
        )

    def test_path(self):
        k1 = caps390.Key("algo", b"digest")

        self.assertEqual(
            k1.path,
            pathlib.Path("caps2") / "algo" / "mr" / "uw" / "ozltoq.xml",
        )

        k1 = caps390.Key("somealgo", b"otherdigest")

        self.assertEqual(
            k1.path,
            pathlib.Path("caps2") / "somealgo" / "n5" / "2g" /
            "qzlsmruwozltoq.xml",
        )

    def test_path_urlencodes_algo(self):
        quote = unittest.mock.Mock(wraps=urllib.parse.quote)
        k = caps390.Key("/.$?", b"digest")

        with unittest.mock.patch("urllib.parse.quote", new=quote):
            path = k.path

        quote.assert_called_once_with("/.$?", safe="")

        self.assertEqual(
            path,
            pathlib.Path("caps2") / "%2F.%24%3F" / "mr" / "uw" / "ozltoq.xml",
        )

    def test_verify_consults__calculate_hash(self):
        k = caps390.Key(
            unittest.mock.sentinel.key_algo,
            unittest.mock.sentinel.key_digest,
        )

        info = unittest.mock.Mock(
            spec=aioxmpp.disco.xso.InfoQuery
        )

        with contextlib.ExitStack() as stack:
            _calculate_hash = stack.enter_context(
                unittest.mock.patch(
                    "aioxmpp.entitycaps.caps390._calculate_hash"
                )
            )
            _calculate_hash.return_value = unittest.mock.sentinel.key_digest

            _get_hash_input = stack.enter_context(
                unittest.mock.patch(
                    "aioxmpp.entitycaps.caps390._get_hash_input"
                )
            )

            self.assertTrue(
                k.verify(info)
            )

            _get_hash_input.assert_called_once_with(
                info,
            )

            _calculate_hash.assert_called_once_with(
                unittest.mock.sentinel.key_algo,
                _get_hash_input(),
            )

    def test_verify_can_use_precalculated_hash_input(self):
        k = caps390.Key(
            unittest.mock.sentinel.key_algo,
            unittest.mock.sentinel.key_digest,
        )

        hash_input = unittest.mock.Mock(
            spec=bytes
        )

        with contextlib.ExitStack() as stack:
            _calculate_hash = stack.enter_context(
                unittest.mock.patch(
                    "aioxmpp.entitycaps.caps390._calculate_hash"
                )
            )
            _calculate_hash.return_value = unittest.mock.sentinel.key_digest

            _get_hash_input = stack.enter_context(
                unittest.mock.patch(
                    "aioxmpp.entitycaps.caps390._get_hash_input"
                )
            )

            self.assertTrue(
                k.verify(hash_input)
            )

            _get_hash_input.assert_not_called()

            _calculate_hash.assert_called_once_with(
                unittest.mock.sentinel.key_algo,
                hash_input,
            )

    def test_verify_detects_mismatch_with_precalculated_hash_input(self):
        k = caps390.Key(
            unittest.mock.sentinel.key_algo,
            unittest.mock.sentinel.key_digest,
        )

        hash_input = unittest.mock.Mock(
            spec=bytes
        )

        with contextlib.ExitStack() as stack:
            _calculate_hash = stack.enter_context(
                unittest.mock.patch(
                    "aioxmpp.entitycaps.caps390._calculate_hash"
                )
            )
            _calculate_hash.return_value = unittest.mock.sentinel.other_digest

            _get_hash_input = stack.enter_context(
                unittest.mock.patch(
                    "aioxmpp.entitycaps.caps390._get_hash_input"
                )
            )

            self.assertFalse(
                k.verify(hash_input)
            )

            _get_hash_input.assert_not_called()

            _calculate_hash.assert_called_once_with(
                unittest.mock.sentinel.key_algo,
                hash_input,
            )

    def test_verify_returns_false_if_digest_mismatches(self):
        k = caps390.Key(
            unittest.mock.sentinel.key_algo,
            unittest.mock.sentinel.key_digest,
        )

        info = unittest.mock.Mock(
            spec=aioxmpp.disco.xso.InfoQuery
        )

        with contextlib.ExitStack() as stack:
            _calculate_hash = stack.enter_context(
                unittest.mock.patch(
                    "aioxmpp.entitycaps.caps390._calculate_hash"
                )
            )
            _calculate_hash.return_value = unittest.mock.sentinel.other_digest

            _get_hash_input = stack.enter_context(
                unittest.mock.patch(
                    "aioxmpp.entitycaps.caps390._get_hash_input"
                )
            )

            self.assertFalse(
                k.verify(info)
            )

            _get_hash_input.assert_called_once_with(
                info,
            )

            _calculate_hash.assert_called_once_with(
                unittest.mock.sentinel.key_algo,
                _get_hash_input(),
            )

    def test_verify_small_input(self):
        k = caps390.Key(
            "sha-256",
            caps390._calculate_hash(
                "sha-256",
                caps390._get_hash_input(TEST_SMALL),
            ),
        )

        self.assertTrue(
            k.verify(TEST_SMALL)
        )

        self.assertFalse(
            k.verify(TEST_LARGE)
        )

    def test_verify_small_input_precalculated(self):
        hash_input_small = caps390._get_hash_input(TEST_SMALL)
        hash_input_large = caps390._get_hash_input(TEST_LARGE)

        k = caps390.Key(
            "sha-256",
            caps390._calculate_hash(
                "sha-256",
                hash_input_small,
            ),
        )

        self.assertTrue(
            k.verify(hash_input_small)
        )

        self.assertFalse(
            k.verify(hash_input_large)
        )

    def test_verify_large_input(self):
        k = caps390.Key(
            "sha-256",
            caps390._calculate_hash(
                "sha-256",
                caps390._get_hash_input(TEST_LARGE),
            ),
        )

        self.assertTrue(
            k.verify(TEST_LARGE)
        )

        self.assertFalse(
            k.verify(TEST_SMALL)
        )

    def test_verify_large_input_precalculated(self):
        hash_input_small = caps390._get_hash_input(TEST_SMALL)
        hash_input_large = caps390._get_hash_input(TEST_LARGE)

        k = caps390.Key(
            "sha-256",
            caps390._calculate_hash(
                "sha-256",
                hash_input_large,
            ),
        )

        self.assertTrue(
            k.verify(hash_input_large)
        )

        self.assertFalse(
            k.verify(hash_input_small)
        )


class TestImplementation(unittest.TestCase):
    def setUp(self):
        self.algorithms = [
            unittest.mock.sentinel.algo1,
            unittest.mock.sentinel.algo2
        ]
        self.i = caps390.Implementation(self.algorithms)

    def test_extract_keys_returns_empty_if_caps_is_None(self):
        presence = unittest.mock.Mock(["xep0390_caps"])
        presence.xep0390_caps = None

        self.assertSetEqual(
            set(self.i.extract_keys(presence)),
            set(),
        )

    def test_extract_keys_creates_Key_objects_from_digests_and_checks_for_support(self):  # NOQA
        presence = unittest.mock.Mock(["xep0390_caps"])
        presence.xep0390_caps.digests = {
            unittest.mock.sentinel.palgo1: unittest.mock.sentinel.pdigest1,
            unittest.mock.sentinel.palgo2: unittest.mock.sentinel.pdigest2,
            unittest.mock.sentinel.palgo3: unittest.mock.sentinel.pdigest3,
        }

        def is_algo_supported_impl(algo):
            return algo != unittest.mock.sentinel.palgo2

        with contextlib.ExitStack() as stack:
            is_algo_supported = stack.enter_context(
                unittest.mock.patch("aioxmpp.hashes.is_algo_supported")
            )
            is_algo_supported.side_effect = is_algo_supported_impl

            result = set(self.i.extract_keys(presence))

        self.assertCountEqual(
            is_algo_supported.mock_calls,
            [
                unittest.mock.call(unittest.mock.sentinel.palgo1),
                unittest.mock.call(unittest.mock.sentinel.palgo2),
                unittest.mock.call(unittest.mock.sentinel.palgo3),
            ]
        )

        self.assertSetEqual(
            result,
            {
                caps390.Key(unittest.mock.sentinel.palgo1,
                            unittest.mock.sentinel.pdigest1),
                caps390.Key(unittest.mock.sentinel.palgo3,
                            unittest.mock.sentinel.pdigest3),
            }
        )

    def test_put_keys_inserts_keys_into_presence(self):
        keys = [
            caps390.Key(unittest.mock.sentinel.palgo1,
                        unittest.mock.sentinel.pdigest1),
            caps390.Key(unittest.mock.sentinel.palgo2,
                        unittest.mock.sentinel.pdigest2),
        ]

        presence = unittest.mock.Mock(["xep0390_caps"])
        presence.xep0390_caps = None

        self.i.put_keys(keys, presence)

        self.assertIsInstance(
            presence.xep0390_caps,
            caps_xso.Caps390,
        )

        self.assertDictEqual(
            presence.xep0390_caps.digests,
            {
                unittest.mock.sentinel.palgo1: unittest.mock.sentinel.pdigest1,
                unittest.mock.sentinel.palgo2: unittest.mock.sentinel.pdigest2,
            }
        )

    def test_calculate_keys_efficiently_builds_hashes_for_given_algorithms(self):  # NOQA
        def generate_digests(algo, _):
            assert isinstance(algo, type(unittest.mock.sentinel.foo))
            return getattr(unittest.mock.sentinel, "{}_digest".format(
                algo.name
            ))

        with contextlib.ExitStack() as stack:
            _calculate_hash = stack.enter_context(
                unittest.mock.patch(
                    "aioxmpp.entitycaps.caps390._calculate_hash"
                )
            )
            _calculate_hash.side_effect = generate_digests

            _get_hash_input = stack.enter_context(
                unittest.mock.patch(
                    "aioxmpp.entitycaps.caps390._get_hash_input"
                )
            )
            _get_hash_input.return_value = unittest.mock.sentinel.hash_input

            result = set(self.i.calculate_keys(unittest.mock.sentinel.info))

        _get_hash_input.assert_called_once_with(
            unittest.mock.sentinel.info
        )

        self.assertCountEqual(
            _calculate_hash.mock_calls,
            [
                unittest.mock.call(algo, unittest.mock.sentinel.hash_input)
                for algo in self.algorithms
            ]
        )

        self.assertSetEqual(
            result,
            {
                caps390.Key(unittest.mock.sentinel.algo1,
                            unittest.mock.sentinel.algo1_digest),
                caps390.Key(unittest.mock.sentinel.algo2,
                            unittest.mock.sentinel.algo2_digest),
            }
        )



########################################################################
# File name: xso.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import enum

import aioxmpp.xso as xso
import aioxmpp.stanza as stanza

from aioxmpp.utils import namespaces


namespaces.xep0085 = "http://jabber.org/protocol/chatstates"


class ChatState(enum.Enum):
    """
    Enumeration of the chat states defined by :xep:`0085`:

    .. attribute:: ACTIVE

    .. attribute:: COMPOSING

    .. attribute:: PAUSED

    .. attribute:: INACTIVE

    .. attribute:: GONE
    """
    ACTIVE = (namespaces.xep0085, "active")
    COMPOSING = (namespaces.xep0085, "composing")
    PAUSED = (namespaces.xep0085, "paused")
    INACTIVE = (namespaces.xep0085, "inactive")
    GONE = (namespaces.xep0085, "gone")


stanza.Message.xep0085_chatstate = xso.ChildTag(ChatState, allow_none=True)



########################################################################
# File name: test___init__.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import unittest

import aioxmpp.entitycaps
import aioxmpp.entitycaps.service
import aioxmpp.entitycaps.xso


class TestExports(unittest.TestCase):
    def test_exports(self):
        self.assertIs(aioxmpp.entitycaps.EntityCapsService,
                      aioxmpp.entitycaps.service.EntityCapsService)
        self.assertIs(aioxmpp.entitycaps.Cache,
                      aioxmpp.entitycaps.service.Cache)
        self.assertIs(aioxmpp.entitycaps.Service,
                      aioxmpp.entitycaps.service.EntityCapsService)
        self.assertIs(aioxmpp.EntityCapsService,
                      aioxmpp.entitycaps.service.EntityCapsService)



########################################################################
# File name: test_statemachine.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import asyncio
import functools
import unittest

from enum import Enum

import aioxmpp.statemachine as statemachine

from aioxmpp.testutils import run_coroutine


@functools.total_ordering
class States(Enum):
    STATE1 = 1
    STATE2 = 2
    STATE3 = 3
    STATE4 = 4

    def __lt__(self, other):
        return self.value < other.value


class TestOrderedStateSkipped(unittest.TestCase):
    def test_is_value_error(self):
        self.assertTrue(issubclass(
            statemachine.OrderedStateSkipped,
            ValueError
        ))

    def test_init(self):
        exc = statemachine.OrderedStateSkipped(States.STATE1)
        self.assertRegex(
            str(exc),
            r"state [^ ]+ has been skipped"
        )
        self.assertIs(exc.skipped_state, States.STATE1)


class TestOrderedStateMachine(unittest.TestCase):
    def setUp(self):
        self.loop = asyncio.get_event_loop()
        self.osm = statemachine.OrderedStateMachine(
            States.STATE1,
            loop=self.loop)

    def tearDown(self):
        del self.osm
        del self.loop

    def test_init(self):
        osm = statemachine.OrderedStateMachine(
            States.STATE1)
        self.assertEqual(
            States.STATE1,
            osm.state
        )

    def test_init_rejects_unordered_state_type(self):
        class OtherStates(Enum):
            FOO = 1
            BAR = 2

        with self.assertRaisesRegex(TypeError,
                                    "states must be ordered"):
            statemachine.OrderedStateMachine(OtherStates.FOO)

    def test_wait_for_at_least(self):
        state_tasks = {
            state: asyncio.ensure_future(self.osm.wait_for_at_least(state),
                                         loop=self.loop)
            for state in States
        }

        run_coroutine(asyncio.sleep(0))

        self.osm.state = States.STATE3

        run_coroutine(asyncio.sleep(0.01))

        for state, task in state_tasks.items():
            if state <= States.STATE3:
                self.assertTrue(task.done(), state)
                self.assertIsNone(task.result())
            else:
                self.assertFalse(task.done(), state)

    def test_wait_for_at_least_checks_current_state(self):
        state_tasks = {
            state: asyncio.ensure_future(self.osm.wait_for_at_least(state),
                                         loop=self.loop)
            for state in States
        }

        self.osm.state = States.STATE3

        run_coroutine(asyncio.sleep(0))

        for state, task in state_tasks.items():
            if state <= States.STATE3:
                self.assertTrue(task.done(), state)
                self.assertIsNone(task.result())
            else:
                self.assertFalse(task.done(), state)

    def test_wait_for_at_least_can_be_cancelled(self):
        state_tasks = {
            state: asyncio.ensure_future(self.osm.wait_for_at_least(state),
                                         loop=self.loop)
            for state in States
        }

        run_coroutine(asyncio.sleep(0))

        for task in state_tasks.values():
            task.cancel()

        run_coroutine(asyncio.sleep(0))

        self.osm.state = States.STATE3

        run_coroutine(asyncio.sleep(0))

    def test_state_rejects_rewinding(self):
        self.osm.state = States.STATE2
        with self.assertRaisesRegex(
                ValueError,
                "cannot rewind OrderedStateMachine"):
            self.osm.state = States.STATE1
        self.osm.state = States.STATE2

    def test_wait_for(self):
        state_tasks = {
            state: asyncio.ensure_future(self.osm.wait_for(state),
                                         loop=self.loop)
            for state in States
        }

        run_coroutine(asyncio.sleep(0))

        self.osm.state = States.STATE3

        run_coroutine(asyncio.sleep(0.01))

        for state, task in state_tasks.items():
            # STATE1 triggers immediately, it is the current state when the
            # coroutine is called
            if state in [States.STATE3, States.STATE1]:
                self.assertTrue(task.done(), state)
                self.assertIsNone(task.result())
            elif state < States.STATE3:
                self.assertTrue(task.done(), state)
                self.assertIsInstance(
                    task.exception(),
                    statemachine.OrderedStateSkipped,
                    state
                )
            else:
                self.assertFalse(task.done(), state)

    def test_wait_for_checks_immediately(self):
        state_tasks = {
            state: asyncio.ensure_future(self.osm.wait_for(state),
                                         loop=self.loop)
            for state in States
        }

        self.osm.state = States.STATE3

        run_coroutine(asyncio.sleep(0.01))

        for state, task in state_tasks.items():
            if state == States.STATE3:
                self.assertTrue(task.done(), state)
                self.assertIsNone(task.result())
            elif state < States.STATE3:
                self.assertTrue(task.done(), state)
                self.assertIsInstance(
                    task.exception(),
                    statemachine.OrderedStateSkipped,
                    state
                )
            else:
                self.assertFalse(task.done(), state)

    def test_rewind_allows_going_back(self):
        self.osm.state = States.STATE3
        self.osm.rewind(States.STATE2)
        self.assertEqual(
            States.STATE2,
            self.osm.state
        )

    def test_rewind_rejects_going_forward(self):
        self.osm.state = States.STATE3
        with self.assertRaisesRegex(ValueError,
                                    "cannot forward using rewind"):
            self.osm.rewind(States.STATE4)

    def test_wait_for_at_least_fails_if_state_is_unreachable(self):
        self.osm.state
